"""
Data structures for Spatial Queries, suah as KNN, within radius searches, etc.. These are good for small and medium sized data,
and data of relatively small dimension (<30).
"""

from __future__ import annotations

import numpy as np
from polars_ds._polars_ds import PyKDT
from typing import List, Tuple
from .typing import KdtDistance

import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:  # 3.10, 3.9, 3.8
    from typing_extensions import Self


class KDTree:
    """
    A Kdtree. This copies data. It is not recommended to use Kdtree on data with dimension > 30. Query speed
    for higher dimensional data is not much better than brute force.
    """

    def __init__(self, X: np.ndarray, distance: KdtDistance = "sql2"):
        """
        Constructs a Kdtree by copying data from a NumPy matrix, where each row is a record with numbers
        in the row being the features.

        The rows will be indexed by [0..row_count(X)).

        Parameters
        ----------
        X
            The data matrix. The rows will be the points.
        distance
            One of 'l1', 'l2', 'sql2', 'inf'
        """
        if X.ndim != 2:
            raise ValueError("Input `X` is not a matrix.")
        if distance not in ("l1", "l2", "sql2", "inf"):
            raise ValueError(f"The given distance `{distance}` is not supported.")

        if X.flags["C_CONTIGUOUS"]:
            self.kdt = PyKDT(X, distance=distance)
        else:
            import warnings

            warnings.warn(
                "Input matrix is not c-contiguous, a copy is made internally and will take up more memory temporarily.",
                stacklevel=2,
            )
            self.kdt = PyKDT(np.ascontiguousarray(X), distance=distance)

    def __len__(self) -> int:
        return self.kdt.count()

    def count(self) -> int:
        return self.kdt.count()

    def add(self, X: np.ndarray) -> Self:
        """
        Appends the new data the tree. The new data will be index by [len(self)..len(self)+row_count(X)).

        Parameters
        ----------

        """
        if X.ndim != 2:
            raise ValueError("Input `X` is not a matrix.")

        self.kdt.add(X)
        return self

    def knn(
        self,
        X: np.ndarray,
        k: int,
        epsilon: float = 0.0,
        max_dist_bound: float = 9999.0,
        parallel: bool = False,
    ) -> (np.ndarray, np.ndarray):
        """
        Returns the K-nearest Neighbors and the distances in two matrices. The first
        is their indices and the second is the corresponding distances.

        Note that if max_dist_bound is low, it is possible that you get < k nearest neighbors.
        In that case, The output matrices will still be the right size but values will be padded. u32::MAX will
        be the index, which is not a valid index in almost all cases and the corresponding distance will
        be inf. You may detect this by ignoring indices > len(tree).

        Parameters
        ----------
        X
            The data matrix (points) to query the KNN for
        k
            Must be a positive int >= 1.
        epsilon
            If non-zero, then it is possible to miss a neighbor within epsilon distance.
        max_dist_bound
            Do not include neighbors beyond this distance.
        parallel
            Whether to run this in parallel or not
        """

        if X.ndim != 2:
            raise ValueError("Input `X` is not a matrix.")

        return self.kdt.knn(
            X, k, epsilon=abs(epsilon), max_dist_bound=abs(max_dist_bound), parallel=parallel
        )

    def within(
        self,
        X: np.ndarray,
        r: float | np.ndarray,
        sort: bool = False,
        return_dist: bool = False,
        parallel: bool = False,
    ) -> List[List[int]] | List[List[Tuple[int, float]]]:
        """
        Returns all points in the tree within radius r from points in X. For each point in X,
        this returns a list of indices of points in the tree that are within the distance from the point.
        If return_dist is true, for each point this will return a list of (int, float) tuple, which
        represents (index, distance).

        Parameters
        ----------
        X
            The data matrix (points) to query the KNN for
        r
            The radius. Either a scalar float, or a 1d array with len = row_count(X).
        sort
            Whether the result in each list should be sorted by distance. If true, this will take a bit longer to run.
        return_dist
            Whether to return only indices or (indices, distance) tuples for each row
        parallel
            Whether to evaluate this in parallel
        """
        if X.ndim != 2:
            raise ValueError("Input `X` is not a matrix.")

        if isinstance(r, (int, float)):
            rad = float(r)
            if return_dist:
                return self.kdt.within_with_dist(X, rad, sort, parallel)
            else:
                return self.kdt.within_idx_only(X, rad, sort, parallel)
        elif isinstance(r, np.ndarray):
            rad = np.ravel(r)
            if len(rad) == 1:
                self.within(X, rad[0], sort, parallel)
            else:
                if return_dist:
                    return self.kdt.within_with_dist_vec_r(X, rad, sort, parallel)
                else:
                    return self.kdt.within_idx_only_vec_r(X, rad, sort, parallel)
        else:
            raise ValueError("Input `r` has to be a scalar or a NumPy array.")

    def within_count(
        self, X: np.ndarray, r: float | np.ndarray, parallel: bool = False
    ) -> List[int]:
        """
        Returns the count of points in the tree within radius r from points in X. This is faster
        than calling within and then count the length of each output.

        Parameters
        ----------
        X
            The data matrix (points) to query the KNN for
        r
            The radius. Either a scalar float, or a 1d array with len = row_count(X).
        parallel
            Whether to evaluate this in parallel
        """
        if X.ndim != 2:
            raise ValueError("Input `X` is not a matrix.")

        if isinstance(r, (int, float)):
            return self.kdt.within_count(X, float(r), parallel)
        elif isinstance(r, np.ndarray):
            rad = np.ravel(r)
            if len(rad) == 1:
                self.within_count(X, rad[0], parallel)
            else:
                return self.kdt.within_count_vec_r(X, rad, parallel)
        else:
            raise ValueError("Input `r` has to be a scalar or a NumPy array.")



"""Time series features."""

from __future__ import annotations

import math
import polars as pl
from typing import Iterable, Literal

# Internal dependencies
from polars_ds.typing import Distance, NullPolicy
from polars_ds._utils import pl_plugin, to_expr

__all__ = [
    "query_mid_range",
    "query_abs_energy",
    "query_symm_ratio",
    "query_mean_abs_change",
    "query_cv",
    "query_mean_n_abs_max",
    "query_longest_streak",
    "query_avg_streak",
    "query_streak",
    "query_count_uniques",
    "query_range_count",
    "query_lempel_ziv",
    "query_first_digit_cnt",
    "query_c3_stats",
    "query_cid_ce",
    "query_time_reversal_asymmetry_stats",
    "query_entropy",
    "query_approx_entropy",
    "query_sample_entropy",
    "query_knn_entropy",
    "query_approx_entropy",
    "query_cond_entropy",
    "query_copula_entropy",
    "query_cond_indep",
    "query_transfer_entropy",
    "query_permute_entropy",
    "query_similar_count",
    "query_ar_coeffs",
    "query_auto_corr",
    # "query_psd",
]

#################################################
# Time series features, some from tsfresh       #
#################################################


def query_mid_range(x: str | pl.Expr) -> pl.Expr:
    """
    A shorthand for (pl.col(x).max() - pl.col(x).min()) / 2.
    """
    xx = to_expr(x)
    return (xx.max() - xx.min()) / 2


def query_symm_ratio(x: str | pl.Expr) -> pl.Expr:
    """
    Returns the symmetric ratio: |mean - median| / (max - min). Note the closer to 0 this value is,
    the more symmetric the series is.
    """
    y = to_expr(x)
    return (y.mean() - y.median()).abs() / (y.max() - y.min())


def query_abs_energy(x: str | pl.Expr) -> pl.Expr:
    """
    Absolute energy is defined as Sum(x_i^2).
    """
    y = to_expr(x)
    return y.dot(y)


def query_mean_abs_change(x: str | pl.Expr) -> pl.Expr:
    """
    Returns the mean of all successive differences |X_i - X_i-1|
    """
    return to_expr(x).diff(null_behavior="drop").abs().mean()


def query_mean_n_abs_max(x: str | pl.Expr, n_maxima: int) -> pl.Expr:
    """
    Returns the average of the top `n_maxima` of |x|.
    """
    if n_maxima <= 0:
        raise ValueError("The number of maxima should be > 0.")
    return to_expr(x).abs().top_k(n_maxima).mean()


def query_cv(x: str | pl.Expr, ddof: int = 1) -> pl.Expr:
    """
    Returns the coefficient of variation for the variable. This is a shorthand for std / mean.

    Parameters
    ----------
    x
        The variable
    ddof
        The delta degree of frendom used in std computation
    """
    xx = to_expr(x)
    return xx.std(ddof=ddof) / xx.mean()


def query_count_uniques(x: str | pl.Expr) -> pl.Expr:
    """
    Returns the count of unique values.
    """
    return to_expr(x).is_unique().sum()


def query_range_count(x: str | pl.Expr, lower: float, upper: float) -> pl.Expr:
    """
    Returns the number of values inside [`lower`, `upper`].
    """
    return to_expr(x).is_between(lower_bound=lower, upper_bound=upper).sum()


def query_longest_streak(where: str | pl.Expr) -> pl.Expr:
    """
    Finds the longest streak length where the condition `where` is true.

    Note: the query is still runnable when `where` doesn't represent boolean column / boolean expressions.
    However, if that is the case the answer will not be easily interpretable.

    Parameters
    ----------
    where
        If where is string, the string must represent the name of a string column. If where is
        an expression, the expression must evaluate to some boolean expression.
    """

    if isinstance(where, str):
        condition = pl.col(where)
    else:
        condition = where

    y = condition.rle().struct.rename_fields(
        ["len", "value"]
    )  # POLARS V1 rename fields can be removed when polars hit v1.0
    return (
        y.filter(y.struct.field("value"))
        .struct.field("len")
        .max()
        .fill_null(0)
        .alias("longest_streak")
    )


def query_avg_streak(where: str | pl.Expr) -> pl.Expr:
    """
    Finds the average streak length where the condition `where` is true. The average is taken on
    the true set.

    Note: the query is still runnable when `where` doesn't represent boolean column / boolean expressions.
    However, if that is the case the answer will not be easily interpretable.

    Parameters
    ----------
    where
        If where is string, the string must represent the name of a string column. If where is
        an expression, the expression must evaluate to some boolean expression.
    """

    if isinstance(where, str):
        condition = pl.col(where)
    else:
        condition = where

    y = condition.rle().struct.rename_fields(
        ["len", "value"]
    )  # POLARS V1 rename fields can be removed when polars hit v1.0
    return (
        y.filter(y.struct.field("value"))
        .struct.field("len")
        .mean()
        .fill_null(0)
        .alias("avg_streak")
    )


def query_streak(where: str | pl.Expr) -> pl.Expr:
    """
    Finds the streak length where the condition `where` is true. This returns a full column of streak lengths.

    Note: the query is still runnable when `where` doesn't represent boolean column / boolean expressions.
    However, if that is the case the answer will not be easily interpretable.

    Parameters
    ----------
    where
        If where is string, the string must represent the name of a boolean column. If where is
        an expression, the expression must evaluate to some boolean series.
    """

    if isinstance(where, str):
        condition = pl.col(where)
    else:
        condition = where

    y = condition.rle().struct.rename_fields(
        ["len", "value"]
    )  # POLARS V1 rename fields can be removed when polars hit v1.0
    return y.struct.field("len").alias("streak_len")


def query_first_digit_cnt(var: str | pl.Expr) -> pl.Expr:
    """
    Finds the first digit count in the data. This is closely related to Benford's law,
    which states that the the first digits (1-9) follow a certain distribution.

    The output is a single element column of type list[u32]. The first value represents the count of 1s
    that are the first digit, the second value represents the count of 2s that are the first digit, etc.

    E.g. first digit of 12 is 1, of 0.0312 is 3. For integers, it is possible to have value = 0, and this
    will not be counted as a first digit.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Benford%27s_law
    """
    return pl_plugin(
        symbol="pl_benford_law",
        args=[to_expr(var)],
        returns_scalar=True,
    )


def query_benford(var: str | pl.Expr) -> pl.Expr:
    """
    Finds the first digit counts which is used in Benford's law. This is an alias to
    `query_first_digit_cnt`.
    """
    return query_first_digit_cnt(var)


def query_similar_count(
    query: Iterable[float],
    target: str | pl.Expr,
    threshold: float,
    metric: Literal["sql2", "sqzl2"] = "sqzl2",
    parallel: bool = False,
    return_ratio: bool = False,
) -> pl.Expr:
    """
    Given a query subsequence, find the number of same-sized subsequences (windows) in target
    series that have distance < threshold from it.

    Note: If target is largely null, errors may occur. If metric is sqzl2, a mininum variance
    of 1e-10 is applied to all variance calculations to avoid division by 0.

    Parameters
    ----------
    query
        The query subsequence. Must not contain nulls.
    target
        The target time series.
    threshold
        The distance threshold
    metric
        Either 'sql2' or 'sqzl2', which stands for squared l2 and squared z-normalized l2.
    parallel
        Only applies when method is `direct`. Whether to compute the convulotion in parallel. Note that this may not
        have the expected performance when you are in group_by or other parallel context already. It is recommended
        to use this in select/with_columns context, when few expressions are being run at the same time.
    return_ratio
        If true, return # of similar subseuqnces / total number of subsequences.
    """

    q = pl.Series(name="", values=query, dtype=pl.Float64)
    if q.null_count() > 0:
        raise ValueError("Nulls found in the query subsequence.")
    if len(q) <= 1:
        raise ValueError("Length of the query should be > 1.")

    t = to_expr(target)
    kwargs = {"threshold": threshold, "parallel": parallel}
    if metric == "sql2":
        result = pl_plugin(
            symbol="pl_subseq_sim_cnt_l2",
            args=[t.cast(pl.Float64).rechunk(), q],
            kwargs=kwargs,
            returns_scalar=True,
        )
    elif metric == "sqzl2":  # pl_subseq_sim_cnt_zl2
        rolling_mean = t.rolling_mean(window_size=len(q)).slice(len(q) - 1, None)
        rolling_var = pl.max_horizontal(
            t.rolling_var(window_size=len(q)).slice(len(q) - 1, None).fill_nan(1e-10),
            pl.lit(1e-10, dtype=pl.Float64),
        )
        qq = pl.lit(q)
        args = [
            t.cast(pl.Float64).rechunk(),
            ((qq - qq.mean()) / qq.std()).rechunk(),
            rolling_mean.rechunk(),
            rolling_var.rechunk(),
        ]
        result = pl_plugin(
            symbol="pl_subseq_sim_cnt_zl2",
            args=args,
            kwargs=kwargs,
            returns_scalar=True,
        )
    else:
        raise ValueError(f"Unsupported metric {metric}.")

    if return_ratio:
        return result / (t.len() - len(q) + 1)
    return result


def query_lempel_ziv(b: str | pl.Expr, as_ratio: bool = True) -> pl.Expr:
    """
    Computes Lempel Ziv complexity on a boolean column. Null will be mapped to False.

    Parameters
    ----------
    b
        A boolean column
    as_ratio : bool
        If true, return complexity / length.
    """
    x = to_expr(b)
    out = pl_plugin(
        symbol="pl_lempel_ziv_complexity",
        args=[x],
        returns_scalar=True,
    )
    if as_ratio:
        return out / x.len()
    return out


def query_c3_stats(x: str | pl.Expr, lag: int) -> pl.Expr:
    """
    Measure of non-linearity in the time series using c3 statistics.

    Parameters
    ----------
    x : pl.Expr
        Either the name of the column or a Polars expression
    lag : int
        The lag that should be used in the calculation of the feature.

    Reference
    ---------
    https://arxiv.org/pdf/chao-dyn/9909043
    """
    two_lags = 2 * lag
    xx = to_expr(x)
    return ((xx.mul(xx.shift(lag)).mul(xx.shift(two_lags))).sum()).truediv(xx.len() - two_lags)


def query_cid_ce(x: str | pl.Expr, normalize: bool = False) -> pl.Expr:
    """
    Estimates the time series complexity.

    Parameters
    ----------
    x : pl.Expr
        Either the name of the column or a Polars expression
    normalize : bool, optional
        If True, z-normalizes the time-series before computing the feature.
        Default is False.

    Reference
    ---------
    https://www.cs.ucr.edu/~eamonn/Complexity-Invariant%20Distance%20Measure.pdf
    """
    xx = to_expr(x)
    if normalize:
        y = (xx - xx.mean()) / xx.std()
    else:
        y = xx

    z = y - y.shift(-1)
    return z.dot(z).sqrt()


def query_time_reversal_asymmetry_stats(x: str | pl.Expr, n_lags: int) -> pl.Expr:
    """
    Queries the Time Reversal Asymmetry Statistic, which is the average of
    (L^2(x) * L(x) - L(x) * x^2), where L is the lag operator.
    """
    y = to_expr(x)
    one_lag = y.shift(-n_lags)
    two_lag = y.shift(-2 * n_lags)  # Nulls won't be in the mean calculation
    return (one_lag * (two_lag + y) * (two_lag - y)).mean()


def query_auto_corr(x: str | pl.Expr, lag: int, ddof: int = 0, normalize: bool = True) -> pl.Expr:
    """
    Computes the auto correlation with the given lag.

    Parameters
    ----------
    x
        The feature
    lag
        The lag
    ddof
        The ddof for the variance
    normalize
        Whether to normalize the value to [-1, 1] or not.
    """
    xx = to_expr(x)
    if normalize:
        x_m = xx - xx.mean()
        var = xx.var(ddof=ddof)
        n = pl.len()
        n_minus_lag = pl.when(n < lag).then(float("nan")).otherwise(n - lag)
        return x_m.dot(x_m.shift(-lag)) / (n_minus_lag * var)
    else:
        return (xx * xx.shift(-lag)).mean()


def query_ar_coeffs(
    x: str | pl.Expr, lag: int, add_bias: bool = True, null_policy: NullPolicy = "raise"
) -> pl.Expr:
    """
    Computes the autoregressive coefficients for the given lag. The bias/intercept term will be the last value in the
    output.

    Parameters
    ----------
    x
        The feature
    lag
        The lag
    add_bias
        Whether to add a bias/intercept term
    null_policy
        One of "raise", "one", "zero", or a finite numeric string.
    """

    if null_policy not in ("raise", "one", "zero"):
        try:
            import math

            z = float(null_policy)
            if not math.isfinite(z):
                raise
        except Exception:
            raise ValueError(
                "`null_polocy` must be 'raise', 'one', 'zero' or any finite numeric string for AR coefficients."
            )

    if lag <= 0:
        raise ValueError("`lag` must be > 0.")

    from . import lin_reg

    xx = to_expr(x)
    return lin_reg(
        *[xx.shift(i).slice(offset=lag).alias(str(i)) for i in range(1, lag + 1)],
        target=xx.slice(offset=lag),
        add_bias=add_bias,
        null_policy=null_policy,
    )


# # Wrong
# def query_psd(x: str | pl.Expr, window_size:int, overlap_size:int) -> pl.Expr:
#
#     xx = to_expr(x)
#     kwargs = {
#         "window_size": window_size,
#         "overlap_size": overlap_size,
#     }
#     return pl_plugin(
#         symbol="pl_psd",
#         args=[xx],
#         kwargs=kwargs,
#         returns_scalar=True,
#     )

#################################################
# Entropies | Entropy related features          #
#################################################


def query_entropy(x: str | pl.Expr, base: float = math.e, normalize: bool = True) -> pl.Expr:
    """
    Computes the entropy of any discrete column. This is shorthand for x.unique_counts().entropy()

    Parameters
    ----------
    x
        Either a string or a polars expression
    base
        Base for the log in the entropy computation
    normalize
        Normalize if the probabilities don't sum to 1.
    """
    return to_expr(x).unique_counts().entropy(base=base, normalize=normalize)


def query_cond_entropy(x: str | pl.Expr, y: str | pl.Expr) -> pl.Expr:
    """
    Queries the conditional entropy of x on y, aka. H(x|y).

    Parameters
    ----------
    x
        Either a string or a polars expression
    y
        Either a string or a polars expression
    """
    return pl_plugin(
        symbol="pl_conditional_entropy",
        args=[to_expr(x), to_expr(y)],
        returns_scalar=True,
        pass_name_to_apply=True,
    )


def query_sample_entropy(
    ts: str | pl.Expr, ratio: float = 0.2, m: int = 2, parallel: bool = False
) -> pl.Expr:
    """
    Calculate the sample entropy of this column. It is highly
    recommended that the user impute nulls before calling this.

    If NaN/some error is returned/thrown, it is likely that:
    (1) Too little data, e.g. m + 1 > length
    (2) ratio or (ratio * std) is too close to or below 0 or std is null/NaN.

    Parameters
    ----------
    ts : str | pl.Expr
        A time series
    ratio : float
        The tolerance parameter. Default is 0.2.
    m : int
        Length of a run of data. Most common run length is 2.
    parallel : bool
        Whether to run this in parallel or not. This is recommended when you
        are running only this expression, and not in group_by context.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Sample_entropy
    """
    if m <= 1:
        raise ValueError("Input `m` must be > 1.")

    t = to_expr(ts)
    r = ratio * t.std(ddof=0)
    rows = t.len() - m + 1

    data = [r, t.slice(0, length=rows)]
    # See rust code for more comment on why I put m + 1 here.
    data.extend(
        t.shift(-i).slice(0, length=rows).alias(str(i)) for i in range(1, m + 1)
    )  # More errors are handled in Rust
    return pl_plugin(
        symbol="pl_sample_entropy",
        args=data,
        kwargs={
            "k": 0,
            "metric": "inf",
            "parallel": parallel,
        },
        returns_scalar=True,
        pass_name_to_apply=True,
    )


def query_approx_entropy(
    ts: str | pl.Expr,
    m: int,
    filtering_level: float,
    scale_by_std: bool = True,
    parallel: bool = True,
) -> pl.Expr:
    """
    Approximate sample entropies of a time series given the filtering level. It is highly
    recommended that the user impute nulls before calling this.

    If NaN/some error is returned/thrown, it is likely that:
    (1) Too little data, e.g. m + 1 > length
    (2) filtering_level or (filtering_level * std) is too close to 0 or std is null/NaN.

    Parameters
    ----------
    ts : str | pl.Expr
        A time series
    m : int
        Length of compared runs of data. This is `m` in the wikipedia article.
    filtering_level : float
        Filtering level, must be positive. This is `r` in the wikipedia article.
    scale_by_std : bool
        Whether to scale filter level by std of data. In most applications, this is the default
        behavior, but not in some other cases.
    parallel : bool
        Whether to run this in parallel or not. This is recommended when you
        are running only this expression, and not in group_by context.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Approximate_entropy
    """

    if filtering_level <= 0 or m <= 1:
        raise ValueError("Filter level must be positive and m must be > 1.")

    t = to_expr(ts)
    if scale_by_std:
        r: pl.Expr = filtering_level * t.std()
    else:
        r: pl.Expr = pl.lit(filtering_level, dtype=pl.Float64)

    rows = t.len() - m + 1
    data = [r, t.slice(0, length=rows).cast(pl.Float64)]
    # See rust code for more comment on why I put m + 1 here.
    data.extend(
        t.shift(-i).slice(0, length=rows).cast(pl.Float64).alias(str(i)) for i in range(1, m + 1)
    )
    # More errors are handled in Rust
    return pl_plugin(
        symbol="pl_approximate_entropy",
        args=data,
        kwargs={
            "k": 0,
            "metric": "inf",
            "parallel": parallel,
        },
        returns_scalar=True,
        pass_name_to_apply=True,
    )


def query_knn_entropy(
    *features: str | pl.Expr,
    k: int = 3,
    dist: Distance = "l2",
    parallel: bool = False,
) -> pl.Expr:
    """
    Computes KNN entropy among all the rows.

    Note if rows <= k, NaN will be returned.

    Parameters
    ----------
    *features
        Columns used as features
    k
        The number of nearest neighbor to consider. Usually 2 or 3.
    dist : Literal[`l2`, `inf`]
        Note `l2` here has to be `l2` with square root.
    parallel : bool
        Whether to run the distance query in parallel. This is recommended when you
        are running only this expression, and not in group_by context.

    Reference
    ---------
    https://arxiv.org/pdf/1506.06501v1.pdf
    """
    if k <= 0:
        raise ValueError("Input `k` must be > 0.")
    if dist not in ["l2", "inf"]:
        raise ValueError("Invalid metric for KNN entropy.")

    return pl_plugin(
        symbol="pl_knn_entropy",
        args=[to_expr(e).alias(str(i)) for i, e in enumerate(features)],
        kwargs={
            "k": k,
            "metric": dist,
            "parallel": parallel,
            "skip_eval": False,
            "skip_data": False,
        },
        returns_scalar=True,
        pass_name_to_apply=True,
    )


def query_copula_entropy(*features: str | pl.Expr, k: int = 3, parallel: bool = False) -> pl.Expr:
    """
    Estimates Copula Entropy via rank statistics.

    Reference
    ---------
    Jian Ma and Zengqi Sun. Mutual information is copula entropy. Tsinghua Science & Technology, 2011, 16(1): 51-54.
    """
    ranks = [x.rank() / x.len() for x in (to_expr(f) for f in features)]
    return -query_knn_entropy(*ranks, k=k, dist="l2", parallel=parallel)


def query_cond_indep(
    x: str | pl.Expr, y: str | pl.Expr, z: str | pl.Expr, k: int = 3, parallel: bool = False
) -> pl.Expr:
    """
    Computes the conditional independance of `x`  and `y`, conditioned on `z`

    Reference
    ---------
    Jian Ma. Multivariate Normality Test with Copula Entropy. arXiv preprint arXiv:2206.05956, 2022.
    """
    # We can likely optimize this by going into Rust.
    # Here we are
    # (1) computing rank multiple times
    # (2) creating 3 separate kd-trees, and copying the data 3 times. Might just need to copy once.
    xyz = query_copula_entropy(x, y, z, k=k, parallel=parallel)
    yz = query_copula_entropy(y, z, k=k, parallel=parallel)
    xz = query_copula_entropy(x, z, k=k, parallel=parallel)
    return xyz - yz - xz


def query_transfer_entropy(
    x: str | pl.Expr, source: str | pl.Expr, lag: int = 1, k: int = 3, parallel: bool = False
) -> pl.Expr:
    """
    Estimating transfer entropy from `source` to `x` with a lag

    Reference
    ---------
    Jian Ma. Estimating Transfer Entropy via Copula Entropy. arXiv preprint arXiv:1910.04375, 2019.
    """
    if lag < 1:
        raise ValueError("Input `lag` must be >= 1.")

    xx = to_expr(x)
    x1 = xx.slice(0, pl.len() - lag)
    x2 = xx.slice(lag, pl.len() - lag)  # (equivalent to slice(lag, None), but will break in v1.0)
    s = to_expr(source).slice(0, pl.len() - lag)
    return query_cond_indep(x2, s, x1, k=k, parallel=parallel)


def query_permute_entropy(
    ts: str | pl.Expr,
    tau: int = 1,
    n_dims: int = 3,
    base: float = math.e,
) -> pl.Expr:
    """
    Computes permutation entropy.

    Parameters
    ----------
    ts : str | pl.Expr
        A time series
    tau : int
        The embedding time delay which controls the number of time periods between elements
        of each of the new column vectors.
    n_dims : int, > 1
        The embedding dimension which controls the length of each of the new column vectors
    base : float
        The base for log in the entropy computation

    Reference
    ---------
    https://www.aptech.com/blog/permutation-entropy/
    """
    if n_dims <= 1:
        raise ValueError("Input `n_dims` has to be > 1.")
    if tau < 1:
        raise ValueError("Input `tau` has to be >= 1.")

    t = to_expr(ts)
    if tau == 1:  # Fast track the most common use case
        return (
            pl.concat_list(t, *(t.shift(-i) for i in range(1, n_dims)))
            .head(t.len() - n_dims + 1)
            .list.eval(pl.element().arg_sort())
            .value_counts()  # groupby and count, but returns a struct
            .struct.field("count")  # extract the field named "count"
            .entropy(base=base, normalize=True)
        )
    else:
        return (
            pl.concat_list(
                t.gather_every(tau),
                *(t.shift(-i).gather_every(tau) for i in range(1, n_dims)),
            )
            .slice(0, length=(t.len() // tau) + 1 - (n_dims // tau))
            .list.eval(pl.element().arg_sort())
            .value_counts()
            .struct.field("count")
            .entropy(base=base, normalize=True)
        )



"""
This module provides classic ML dataset transforms. Note all functions here are single-use only, meaning
that the data learned (e.g. mean value in mean imputation) will not be preserved. For pipeline usage, which
preserves the learned values and optimizes the transform query, see pipeline.py.
"""

from __future__ import annotations

import polars as pl
import polars.selectors as cs
from typing import List, Dict, Literal

# Internal dependencies
from polars_ds.typing import (
    PolarsFrame,
    SimpleImputeMethod,
    SimpleScaleMethod,
    ExprTransform,
    QuantileMethod,
    EncoderDefaultStrategy,
)
import polars_ds.exprs.num as pds_num
import polars_ds.exprs.expr_linear as lr


def impute(df: PolarsFrame, cols: List[str], method: SimpleImputeMethod = "mean") -> ExprTransform:
    """
    Impute null values in the given columns. This transform will collect if input is lazy.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names
    method
        One of `mean`, `median`, `mode`. If `mode`, a random value will be chosen if there is
        a tie.
    """
    if method == "mean":
        temp = df.lazy().select(pl.col(cols).mean()).collect().row(0)
        return [pl.col(c).fill_null(m) for c, m in zip(cols, temp)]
    elif method == "median":
        temp = df.lazy().select(pl.col(cols).median()).collect().row(0)
        return [pl.col(c).fill_null(m) for c, m in zip(cols, temp)]
    elif method == "mode":
        temp = df.lazy().select(pl.col(cols).mode().list.first()).collect().row(0)
        return [pl.col(c).fill_null(m) for c, m in zip(cols, temp)]
    else:
        raise ValueError(f"Unknown impute method: `{method}`")


def conditional_impute(
    df: PolarsFrame, rules_dict: Dict[str, str | pl.Expr], method: SimpleImputeMethod = "mean"
) -> ExprTransform:
    """
    Conditionally imputes values in the given columns. This transform will collect if input is lazy.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    rules_dict
        Dictionary where keys are column names (must be string), and values are SQL/Polars Conditions
        that when true, those values in the column will be imputed,
        and the value to impute will be learned on the data where the condition is false.
    method
        One of `mean`, `median`, `mode`. If `mode`, a random value will be chosen if there is
        a tie.
    """
    rules_dict = {
        c: (r if isinstance(r, pl.Expr) else pl.sql_expr(r)) for c, r in rules_dict.items()
    }
    cols = list(rules_dict.keys())
    # Learn on the data where the condition is false
    if method == "mean":
        temp = (
            df.lazy()
            .select(*(pl.col(c).filter(rules_dict[c].not_()).mean() for c in rules_dict.keys()))
            .collect()
            .row(0)
        )
        return [
            pl.when(rules_dict[c]).then(m).otherwise(pl.col(c)).alias(c) for c, m in zip(cols, temp)
        ]
    elif method == "median":
        temp = (
            df.lazy()
            .select(*(pl.col(c).filter(rules_dict[c].not_()).median() for c in rules_dict.keys()))
            .collect()
            .row(0)
        )
        return [
            pl.when(rules_dict[c]).then(m).otherwise(pl.col(c)).alias(c) for c, m in zip(cols, temp)
        ]
    elif method == "mode":
        temp = (
            df.lazy()
            .select(
                *(
                    pl.col(c).filter(rules_dict[c].not_()).mode().list.first()
                    for c in rules_dict.keys()
                )
            )
            .collect()
            .row(0)
        )
        return [
            pl.when(rules_dict[c]).then(m).otherwise(pl.col(c)).alias(c) for c, m in zip(cols, temp)
        ]
    else:
        raise ValueError(f"Unknown impute method: `{method}`")


def linear_impute(
    df: PolarsFrame, features: List[str], target: str | pl.Expr, add_bias: bool = False
) -> ExprTransform:
    """
    Imputes the target column by training a simple linear regression using the other features. This will
    cast the target column to f64.

    Note: The linear regression will skip nulls whenever there is a null in the features or in the target.
    Additionally, if NaN or Inf exists in data, the linear regression result may be invalid or an error
    will be thrown. It is recommended to use this only after imputing and dealing with NaN and Infs for
    all feature columns first.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    features
        A list of strings representing column names that will be used as features in the linear regression
    target
        The target column
    add_bias
        Whether to add a bias term to the linear regression
    """
    target_name = df.lazy().select(target).collect_schema().names()[0]
    features_as_expr = [pl.col(f) for f in features]
    target_as_expr = pl.col(target_name)
    temp = (
        df.lazy()
        .select(
            lr.lin_reg(
                *features_as_expr, target=target_as_expr, add_bias=add_bias, null_policy="skip"
            )
        )
        .collect()
    )  # Add streaming config
    coeffs = temp.item(0, 0)  # coeffs is a series
    linear_eq = [f * coeffs[i] for i, f in enumerate(features_as_expr)]
    if add_bias:
        linear_eq.append(pl.lit(coeffs[-1], dtype=pl.Float64))

    return [pl.col(target_name).fill_null(pl.sum_horizontal(linear_eq)).alias(target_name)]


def center(df: PolarsFrame, cols: List[str]) -> ExprTransform:
    """
    Center the given columns so that they will have 0 mean.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names
    """
    means = df.lazy().select(pl.col(cols).mean()).collect().row(0)
    return [pl.col(c) - m for c, m in zip(cols, means)]


def scale(
    df: PolarsFrame,
    cols: List[str],
    method: SimpleScaleMethod = "standard",
) -> ExprTransform:
    """
    Scales values in the given columns. This transform will collect if input is lazy.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names
    method
        One of `standard`, `min_max`, `abs_max`
    """
    if method == "standard":
        temp = (
            df.lazy()
            .select(
                pl.col(cols).mean().name.prefix("mean:"),
                pl.col(cols).std(ddof=0).name.prefix("std:"),
            )
            .collect()
            .row(0)
        )
        n = len(cols)
        return [(pl.col(c) - temp[i]) / temp[i + n] for i, c in enumerate(cols)]
    elif method == "min_max":
        temp = (
            df.lazy()
            .select(
                pl.col(cols).min().name.prefix("min:"),
                pl.col(cols).max().name.prefix("max:"),
            )
            .collect()
            .row(0)
        )
        n = len(cols)
        # If input is constant, this will return all NaNs.
        return [(pl.col(c) - temp[i]) / (temp[n + i] - temp[i]) for i, c in enumerate(cols)]
    elif method == "abs_max":
        temp = (
            df.lazy()
            .select(pl.max_horizontal(pl.col(c).min().abs(), pl.col(c).max().abs()) for c in cols)
            .collect()
            .row(0)
        )
        return [pl.col(c) / m for c, m in zip(cols, temp)]
    else:
        raise ValueError(f"Unknown input method: {method}")


def robust_scale(
    df: PolarsFrame,
    cols: List[str],
    q_low: float = 0.25,
    q_high: float = 0.75,
    method: QuantileMethod = "midpoint",
) -> ExprTransform:
    """
    Like min-max scaling, but scales each column by the quantile value at q1 and q2.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names
    q_low
        The lower quantile value
    q_high
        The higher quantile value
    method
        Method to compute quantile. One of `nearest`, `higher`, `lower`, `midpoint`, `linear`.
    """
    if q_low > 1.0 or q_low < 0.0 or q_high > 1.0 or q_high < 0.0 or q_low >= q_high:
        raise ValueError(
            "Input `q_low` and `q_high` must be between 0 and 1 and q_low must be < than q_high."
        )

    temp = (
        df.lazy()
        .select(
            pl.col(cols).quantile(q_low, interpolation=method).name.prefix("q1:"),
            pl.col(cols).quantile(q_high, interpolation=method).name.prefix("q2:"),
        )
        .collect()
        .row(0)
    )
    n = len(cols)
    return [(pl.col(c) - temp[i]) / (temp[n + i] - temp[i]) for i, c in enumerate(cols)]


def winsorize(
    df: PolarsFrame,
    cols: List[str],
    q_low: float = 0.05,
    q_high: float = 0.95,
    method: QuantileMethod = "nearest",
) -> ExprTransform:
    """
    Learns the lower and upper percentile from the columns, then clip each end at those values.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names
    q_low
        The lower quantile value
    q_high
        The higher quantile value
    method
        Method to compute quantile. One of `nearest`, `higher`, `lower`, `midpoint`, `linear`.
    """
    if q_low > 1.0 or q_low < 0.0 or q_high > 1.0 or q_high < 0.0 or q_low >= q_high:
        raise ValueError(
            "Input `q_low` and `q_high` must be between 0 and 1 and q_low must be < than q_high."
        )

    temp = (
        df.lazy()
        .select(
            pl.col(cols).quantile(q_low, interpolation=method).name.prefix("l"),
            pl.col(cols).quantile(q_high, interpolation=method).name.prefix("u"),
        )
        .collect()
        .row(0)
    )
    n = len(cols)
    return [pl.col(c).clip(temp[i], temp[n + i]) for i, c in enumerate(cols)]


def one_hot_encode(
    df: PolarsFrame, cols: List[str], separator: str = "_", drop_first: bool = False
) -> ExprTransform:
    """
    Find the unique values in the string/categorical columns and one-hot encode them. This will NOT
    consider nulls as one of the unique values. Append a one-hot null indicator if you want to encode nulls.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names. Columns of type != string/categorical will not produce any expression.
    separator
        E.g. if column name is `col` and `a` is an elemenet in it, then the one-hot encoded column will be called
        `col_a` where the separator `_` is used.
    drop_first
        Whether to drop the first distinct value (in terms of str/categorical order). This helps with reducing
        dimension and prevents some issues from linear dependency.
    """

    temp = (
        df.lazy()
        .select(cols)
        .select(
            (cs.string() | cs.categorical())
            .unique()
            .drop_nulls()
            .cast(pl.String)
            .implode()
            .list.sort()
        )
    )
    exprs: list[pl.Expr] = []
    for t in temp.collect().get_columns():
        u: pl.Series = t[0]  # t is a Series which contains a single series, so u is a series
        if len(u) > 1:
            # Need to take care of the case where null == 1 is null. Need only True and False, not null
            exprs.extend(
                pl.col(t.name).eq_missing(u[i]).cast(pl.UInt8).alias(t.name + separator + u[i])
                for i in range(int(drop_first), len(u))
            )

    if len(exprs) == 0:
        raise ValueError(
            "Provided columns either do not exist or are not string/categorical types."
        )

    return exprs


def ordinal_encode(
    df: PolarsFrame,
    cols: List[str],
    unknown_value: float | None = None,
    null_value: float | None = None,
) -> ExprTransform:
    """
    Converts numeric / string columns into integer encodings starting from 0 to n-1, where
    n is the number of distinct values in the column.

    Note: if assigning an integer based on sorting does not make sense for
    your use case, then please directly use `pl.col().replace_strict(mapping)`

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names.
    unknown_value
        What to assign to values not present in training dataset. None means null.
        Must be a number.
    null_value
        What to assign to values that are null. Must be a number.
    """
    temp = (
        df.lazy()
        .select(cols)
        .select(
            (cs.string() | cs.categorical() | cs.numeric() | cs.boolean())
            .unique()
            .drop_nulls()
            .sort()
            .implode()
        )
    )
    exprs: list[pl.Expr] = []
    for t in temp.collect().get_columns():
        # u is a sorted series
        u: pl.Series = t[0]  # old
        new = pl.int_range(0, len(u), eager=True).cast(pl.Float64)
        c = (
            pl.when(pl.col(t.name).is_null())
            .then(pl.lit(null_value, dtype=pl.Float64))
            .otherwise(
                pl.col(t.name).replace_strict(
                    old=u,
                    new=new,
                    default=pl.lit(unknown_value, dtype=pl.Float64),
                    return_dtype=pl.Float64,
                )
            )
            .alias(t.name)
        )

        exprs.append(c)

    return exprs


def rank_hot_encode(
    col: str,
    ranking: List[str],
    default_rank: int | None = None,
) -> ExprTransform:
    """
    Given a ranking, e.g. ["bad", "neutral", "good"], where "bad", "neutral" and "good" are values coming
    from the column `col`, this will return two new columns, the first is ">=neutral", which
    will be 1 for all values in ("neutral", "good") and 0 otherwise, and the second new column is ">=good", which
    will be 1 for all values in ("good") and 0 otherwise.

    This currently only works on string columns.

    Values not in the provided ranking will have -1 in all the new columns.

    Parameters
    ----------
    col
        The name of a single column
    ranking
        A list of string representing the ranking of the values
    default_rank
        Default rank for all null/unseen values
    """

    n_ranks = len(ranking)
    if n_ranks <= 1:
        raise ValueError("Rank hot encoding does not work with single value ranking.")

    number_rank = pl.int_range(0, n_ranks, eager=True, dtype=pl.Int32)
    ranked_expr = pl.col(col).replace_strict(
        old=ranking, new=number_rank, default=default_rank, return_dtype=pl.Int32
    )

    return [
        (ranked_expr >= i).cast(pl.Int8).alias(f"{col}>={c}")
        for i, c in zip(range(1, n_ranks), ranking[1:])
    ]


def _encoder_default_value(
    temp: PolarsFrame,
    default: EncoderDefaultStrategy | float | None,
    target: str | pl.Expr | pl.Series,
) -> float | None:
    """
    Finds the default value for encoders (Target, WOE, IV encoders) for null and unknown values.
    """
    if default is None or isinstance(default, (int, float)):
        return default
    elif isinstance(default, str):
        if default == "null":
            return None
        elif default == "zero":
            return 0.0
        elif default == "mean":
            if isinstance(target, str):
                return temp.lazy().select(pl.col(target).mean()).collect().item(0, 0)
            elif isinstance(target, pl.Expr):
                return temp.lazy().select(target.mean()).collect().item(0, 0)
            elif isinstance(target, pl.Series):
                return target.mean()
            else:
                raise ValueError("Target's type is not supported.")
        else:
            raise ValueError(
                "When input `default` is string, it can only be `mean` or `null` or `zero`."
            )
    else:
        raise ValueError("Invalid type for `default`")


def target_encode(
    df: PolarsFrame,
    cols: List[str],
    /,
    target: str | pl.Expr | pl.Series,
    min_samples_leaf: int = 20,
    smoothing: float = 10.0,
    default: EncoderDefaultStrategy | float | None = "null",
) -> ExprTransform:
    """
    Target encode the given variables. This will overwrite the columns that will be encoded.

    Note: Nulls will always be mapped to the default.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names. Columns of type != string/categorical will not produce any expression.
    target
        The target column
    min_samples_leaf
        A regularization factor
    smoothing
        Smoothing effect to balance categorical average vs prior
    default
        If a new value is encountered during transform (unseen in training dataset), it will be mapped to default.
        If this is a string, it can be `null`, `zero`, or `mean`, where `mean` means map them to the mean of the target.

    Reference
    ---------
    https://contrib.scikit-learn.org/category_encoders/targetencoder.html
    """
    temp = df.lazy()
    valid_cols = temp.select(cols).select(cs.string() | cs.categorical()).collect_schema().names()

    if len(valid_cols) == 0:
        raise ValueError(
            "The provided columns are either not string/categorical type, or are not in df."
        )

    default_value = _encoder_default_value(temp, default=default, target=target)

    temp = temp.select(
        pds_num.target_encode(
            c, target, min_samples_leaf=min_samples_leaf, smoothing=smoothing
        ).implode()
        for c in valid_cols
    ).collect()  # add collect config..
    return [
        # c[0] will be a series of struct because of the implode above.
        pl.col(c.name).replace_strict(
            old=c[0].struct.field("value"), new=c[0].struct.field("to"), default=default_value
        )
        for c in temp.get_columns()
    ]


def woe_encode(
    df: PolarsFrame,
    cols: List[str],
    /,
    target: str | pl.Expr | pl.Series,
    default: EncoderDefaultStrategy | float | None = "null",
) -> ExprTransform:
    """
    Use Weight of Evidence to encode a discrete variable x with respect to target. This assumes x
    is discrete and castable to String. A value of 1 is added to all events/non-events
    (goods/bads) to smooth the computation. This is -1 * output of the package category_encoder's WOEEncoder.

    Note: Nulls will always be mapped to the default.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names. Columns of type != string/categorical will not produce any expression.
    target
        The target column
    default
        If a new value is encountered during transform (unseen in training dataset), it will be mapped to default.
        If this is a string, it can be `null`, `zero`, or `mean`, where `mean` means map them to the mean of the target.

    Reference
    ---------
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
    """
    temp = df.lazy()
    valid_cols = temp.select(cols).select(cs.string() | cs.categorical()).collect_schema().names()

    if len(valid_cols) == 0:
        raise ValueError(
            "The provided columns are either not string/categorical type, or are not in df."
        )

    default_value = _encoder_default_value(temp, default=default, target=target)

    temp = temp.select(
        pds_num.woe_discrete(c, target).implode() for c in valid_cols
    ).collect()  # add collect config..

    return [
        # c[0] will be a series of struct because of the implode above.
        pl.col(c.name).replace_strict(
            old=c[0].struct.field("value"), new=c[0].struct.field("woe"), default=default_value
        )
        for c in temp.get_columns()
    ]


def iv_encode(
    df: PolarsFrame,
    cols: List[str],
    /,
    target: str | pl.Expr | pl.Series,
    default: EncoderDefaultStrategy | float | None = "null",
) -> ExprTransform:
    """
    Use Information Value to encode a discrete variable x with respect to target. This assumes x
    is discrete and castable to String. A value of 1 is added to all events/non-events
    (goods/bads) to smooth the computation.

    Note: Nulls will always be mapped to the default.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names. Columns of type != string/categorical will not produce any expression.
    target
        The target column
    default
        If a new value is encountered during transform (unseen in training dataset), it will be mapped to default.
        If this is a string, it can be `null`, `zero`, or `mean`, where `mean` means map them to the mean of the target.

    Reference
    ---------
    https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
    """
    temp = df.lazy()
    valid_cols = temp.select(cols).select(cs.string() | cs.categorical()).collect_schema().names()

    if len(valid_cols) == 0:
        raise ValueError(
            "The provided columns are either not string/categorical type, or are not in df."
        )

    default_value = _encoder_default_value(temp, default=default, target=target)

    temp = temp.select(
        pds_num.info_value_discrete(c, target, return_sum=False).implode() for c in valid_cols
    ).collect()  # add collect config..
    # POLARS_V1
    return [
        # c[0] will be a series of struct because of the implode above.
        pl.col(c.name).replace_strict(
            old=c[0].struct.field("value"), new=c[0].struct.field("iv"), default=default_value
        )
        for c in temp.get_columns()
    ]

def select_by_std(
    df: PolarsFrame,
    cols: List[str],
    /,
    min_: float,
    max_: float,
) -> ExprTransform:
    """
    Fits and selects from `cols` columns that have standard deviation between `min_` and `max_`.
    Non-numeric columns in `cols` will always be selected.

    Parameters
    ----------
    df
        Either a lazy or an eager dataframe
    cols
        A list of strings representing column names.
    min_
        Min standard deviation to select, inclusive.
    max_ 
        Max standard deviation to select, exclusive
    """
    numeric = df.lazy().select(
        cs.by_name(cols) & cs.numeric()
    ).collect_schema().names()
    std_vals = df.lazy().select(pl.col(numeric).std()).collect().row(0)
    # Do an exclusion instead
    return pl.exclude([c for c, v in zip(numeric, std_vals) if (v < min_ or v > max_)])

def polynomial_features(
    cols: List[str],
    /,
    degree: int,
    interaction_only: bool = False,
) -> ExprTransform:
    """
    Generates polynomial combinations out of the features given, at the given degree.

    Parameters
    ----------
    cols
        A list of strings representing column names.
    degree
        The degree of the polynomial combination
    interaction_only
        It true, only combinations that involve 2 or more variables will be used.
    """
    from itertools import combinations_with_replacement

    if degree <= 1:
        raise ValueError("Degree should be > 1.")

    return list(
        pl.reduce(function=lambda acc, x: acc * x, exprs=list(comb)).alias("*".join(comb))
        for comb in combinations_with_replacement(cols, degree)
        if ((not interaction_only) or len(set(comb)) > 1)
    )



from __future__ import annotations

import polars as pl
import altair as alt
from typing import Iterable, List
from polars._typing import IntoExpr

# Internal dependencies
import polars_ds.sample_and_split as sa
from polars_ds import query_r2, principal_components, query_tpr_fpr, integrate_trapz

alt.data_transformers.enable("vegafusion")

# Plots should never have a title. Title must be editable by the end user
# Interactivity should only be enabled by the end user


def plot_feature_distr(
    *,
    feature: str | Iterable[float],
    n_bins: int = 10,
    density: bool = False,
    show_bad_values: bool = True,
    over: str | None = None,
    df: pl.DataFrame | pl.LazyFrame | None = None,
) -> alt.Chart:
    """
    Plot distribution of the feature with a few statistical details.

    Parameters
    ----------
    df
        Either an eager or lazy Polars Dataframe
    feature
        A string representing a column name
    n_bins
        The max number of bins used for histograms.
    density
        Whether to plot a probability density or not
    show_bad_values
        Whether to show % of bad (null or non-finite) values
    over
        Whether to look at the distribution over another categorical column
    """

    if n_bins <= 2:
        raise ValueError("Input `n_bins` must be > 2.")

    if over is not None and df is None:
        raise ValueError("Input `over` can only be used when df is not None.")

    if isinstance(feature, str):
        if df is None:
            raise ValueError("If `feature` is str, then df cannot be none.")
        feat = feature
        if over is None:
            data = df.lazy().select(pl.col(feat).cast(pl.Float64)).collect()
        else:
            data = df.lazy().select(pl.col(feat).cast(pl.Float64), over).collect()
    else:
        if over is None:
            data = pl.Series(name="feature", values=feature, dtype=pl.Float64).to_frame()
            feat = "feature"
        else:
            raise ValueError("If input `feature` is a Series, then `over` cannot be used.")

    # selection = alt.selection_point(fields=['species'], bind='legend')
    # .filter(pl.col(feat).is_not_null())
    if density:
        if over is None:
            chart = (
                alt.Chart(data)
                .transform_density(
                    feat,
                    as_=[feat, "density"],
                )
                .mark_area()
                .encode(
                    x=f"{feat}:Q",
                    y=alt.Y("density:Q").stack(None),
                )
            )
        else:
            selection = alt.selection_multi(fields=[over], bind="legend")
            chart = (
                alt.Chart(data)
                .transform_density(feat, as_=[feat, "density"], groupby=[over])
                .mark_area()
                .encode(
                    x=f"{feat}:Q",
                    y=alt.Y("density:Q").stack(None),
                    color=over,
                    opacity=alt.condition(selection, alt.value(0.8), alt.value(0.2)),
                )
                .add_params(selection)
            )
    else:
        if over is None:
            chart = (
                alt.Chart(data)
                .mark_bar()
                .encode(
                    alt.X(f"{feat}:Q").bin(maxbins=n_bins).title(feat),
                    y=alt.Y("count()").stack(None),
                )
            )
        else:
            selection = alt.selection_point(fields=[over], bind="legend")
            chart = (
                alt.Chart(data)
                .mark_bar()
                .encode(
                    alt.X(f"{feat}:Q").bin(maxbins=n_bins).title(feat),
                    y=alt.Y("count()").stack(None),
                    color=over,
                    opacity=alt.condition(selection, alt.value(0.8), alt.value(0.2)),
                )
                .add_params(selection)
            )

    if over is None:
        p5, median, mean, p95, min_, max_, cnt, null_cnt, not_finite = data.select(
            p5=pl.col(feat).quantile(0.05),
            median=pl.col(feat).median(),
            mean=pl.col(feat).mean(),
            p95=pl.col(feat).quantile(0.95),
            min=pl.col(feat).min(),
            max=pl.col(feat).max(),
            cnt=pl.len(),
            null_cnt=pl.col(feat).null_count(),
            not_finite=pl.col(feat).is_finite().not_().sum(),
        ).row(0)

        # stats overlay
        df_stats = pl.DataFrame(
            {"names": ["p5", "median", "avg", "p95"], "stats": [p5, median, mean, p95]}
        )

        stats_base = alt.Chart(df_stats)
        stats_chart = stats_base.mark_rule(color="#f086ab").encode(
            x=alt.X("stats").title(""),
            tooltip=[
                alt.Tooltip("names:N", title="Stats"),
                alt.Tooltip("stats:Q", title="Value"),
            ],
        )

        chart = chart + stats_chart
        if show_bad_values:
            df_bad_values = pl.DataFrame(
                {
                    "names": [""],
                    "pcts": [(null_cnt + not_finite) / cnt],
                }
            )

            bad_values_chart = (
                alt.Chart(df_bad_values)
                .mark_bar(opacity=0.7)
                .encode(
                    x=alt.X("pcts:Q", scale=alt.Scale(domain=[0, 1]))
                    .axis(format=".0%")
                    .title("Null or Non-Finite %"),
                    y=alt.Y("names:N").title(""),
                    tooltip=[
                        alt.Tooltip("pcts:Q", title="Null or Non-Finite %"),
                    ],
                )
            )

            return alt.vconcat(chart, bad_values_chart)
        else:
            return chart

    else:  # over is not None
        if show_bad_values:
            df_bad = data.group_by(over).agg(
                pcts=(pl.col(feat).null_count() + pl.col(feat).is_finite().not_().sum()) / pl.len()
            )
            bad_values_chart = (
                alt.Chart(df_bad)
                .mark_bar(opacity=0.7)
                .encode(
                    x=alt.X("pcts:Q", scale=alt.Scale(domain=[0, 1]))
                    .axis(format=".0%")
                    .title("Null or Non-Finite %"),
                    y=alt.Y(f"{over}:N"),
                    tooltip=[
                        alt.Tooltip("pcts:Q", title="Null or Non-Finite %"),
                    ],
                )
            )
            return alt.vconcat(chart, bad_values_chart)
        else:
            return chart


def plot_lin_reg(
    df: pl.DataFrame | pl.LazyFrame,
    x: str,
    target: str,
    add_bias: bool = False,
    weights: str | None = None,
    max_points: int = 20_000,
    show_lin_reg_eq: bool = True,
) -> alt.Chart:
    """
    Plots the linear regression line between x and target.

    Paramters
    ---------
    df
        Either an eager or lazy Polars Dataframe
    x
        The preditive variable
    target
        The target variable
    add_bias
        Whether to add bias in the linear regression
    weights
        Weights for the linear regression
    max_points
        The max number of points to be displayed. Notice that this only affects the number of points
        on the plot. The linear regression will still be fit on the entire dataset.
    show_lin_reg_eq
        Whether to show the linear regression equation at the bottom or not
    """

    to_select = [x, target] if weights is None else [x, target, weights]
    temp = df.lazy().select(*to_select)

    xx = pl.col(x)
    yy = pl.col(target)
    # Although using simple_lin_reg might seem to be able to reduce some code here,
    # it adds complexity because of output type and the r2 query.
    # A little bit of code dup is reasonable.
    if add_bias:
        if weights is None:
            x_mean = xx.mean()
            y_mean = yy.mean()
            beta = (xx - x_mean).dot(yy - y_mean) / (xx - x_mean).dot(xx - x_mean)
            alpha = y_mean - beta * x_mean
        else:
            w = pl.col(weights)
            w_sum = w.sum()
            x_wmean = w.dot(xx) / w_sum
            y_wmean = w.dot(yy) / w_sum
            beta = w.dot((xx - x_wmean) * (yy - y_wmean)) / (w.dot((xx - x_wmean).pow(2)))
            alpha = y_wmean - beta * x_wmean
    else:
        if weights is None:
            beta = xx.dot(yy) / xx.dot(xx)
        else:
            w = pl.col(weights)
            beta = w.dot(xx * yy) / w.dot(xx.pow(2))

        alpha = pl.lit(0, dtype=pl.Float64)

    beta, alpha, r2, length = (
        temp.select(
            beta.alias("beta"),
            alpha.alias("alpha"),
            query_r2(yy, xx * beta + alpha).alias("r2"),
            pl.len(),
        )
        .collect()
        .row(0)
    )

    df_need = temp.select(
        xx,
        yy,
        (xx * beta + alpha).alias("y_pred"),
    )
    # Sample down if len(temp) > max_points
    df_sampled = sa.sample(df_need, value=max_points) if length > max_points else df_need.collect()

    x_title = [x]
    if show_lin_reg_eq:
        if add_bias and alpha > 0:
            reg_info = f"y = {beta:.4f} * x + {round(alpha, 4) if add_bias else ''}, r2 = {r2:.4f}"
        elif add_bias and alpha < 0:
            reg_info = (
                f"y = {beta:.4f} * x - {abs(round(alpha, 4)) if add_bias else ''}, r2 = {r2:.4f}"
            )
        else:
            reg_info = f"y = {beta:.4f} * x, r2 = {r2:.4f}"

        x_title.append(reg_info)

    chart = alt.Chart(df_sampled).mark_point().encode(alt.X(x).scale(zero=False), alt.Y(target))
    return chart + chart.mark_line().encode(
        alt.X(x, title=x_title).scale(zero=False),
        alt.Y("y_pred"),
    )


def plot_pca(
    df: pl.DataFrame | pl.LazyFrame,
    features: List[str],
    by: IntoExpr,
    center: bool = True,
    dim: int = 2,
    filter_by: pl.Expr | None = None,
    max_points: int = 10_000,
    **kwargs,
) -> alt.Chart:
    """
    Creates a scatter plot based on the reduced dimensions via PCA, and color it by `by`.

    Paramters
    ---------
    df
        Either an eager or lazy Polars Dataframe
    features
        List of feature names
    by
        Color the 2-D PCA plot by the values in the column
    center
        Whether to automatically center the features
    dim
        Only 2 principal components plot can be done at this moment.
    filter_by
        A boolean expression
    max_points
        The max number of points to be displayed. If data > this limit, the data will be sampled.
    kwargs
        Anything else that will be passed to Altair encode function
    """
    if len(features) < 2:
        raise ValueError("You must pass >= 2 features.")
    if dim not in (2, 3):
        raise ValueError("Dim must be 2 or 3.")

    frame = df if filter_by is None else df.filter(filter_by)

    temp = frame.select(principal_components(*features, center=center, k=dim).alias("pc"), by)
    df_plot = sa.sample(temp, value=max_points).unnest("pc")

    if dim == 2:
        return (
            alt.Chart(df_plot).mark_circle(size=60).encode(x="pc1", y="pc2", color=by, **kwargs)
        )  # .interactive()
    else:  # 3d
        raise NotImplementedError


def plot_prob_calibration(
    *,
    target: Iterable[int],
    score: pl.Series | None = None,
    name: str | None = None,
    scores: List[pl.Series] | None = None,
    names: List[str] | None = None,
    n_bins: int = 10,
) -> alt.Chart:
    """
    Plots probability calibration of score(s) with respect to the binary target.

    Parameters
    ----------
    target
        The target binary varialbe
    score
        The probability score values
    name
        The name of the probability score values
    scores
        If score is None, and scores is a list of probability scores, this will
        generate a plot with all probability calibrations.
    names
        If scores is population, this must be a list of corresponding score names.
    n_bins
        N quantile bins for the score(s).
    """

    if score is not None:
        if name is None:
            raise ValueError("If `score` is not None, then `name` must not be none.")
        else:
            new_dict = {name: pl.Series(values=score)}

    else:  # score is None
        if (scores is None) or (names is None):
            raise ValueError("If `score` is None, then `scores` and `names` must be populated.")

        if hasattr(scores, "__len__") and (hasattr(names, "__len__")):
            if len(scores) != len(names):
                raise ValueError("Input `scores` and `names` must have the same length.")

            new_dict = {n: pl.Series(values=s) for n, s in zip(names, scores)}
        else:
            raise ValueError("Input `scores` and `names` must be iterables with a length.")

    target_series = pl.Series(name="__actual__", values=target)

    if any(len(s) != len(target_series) for s in new_dict.values()):
        raise ValueError("All input `score(s)` and `target` must have the same length.")

    new_dict["__actual__"] = target_series

    df = pl.from_dict(new_dict)
    perfect_line = pl.int_range(1, 100, step=5, eager=True) / 100
    df_line = pl.DataFrame(
        {"mean_predicted_prob": perfect_line, "fraction_of_positives": perfect_line}
    ).with_columns(score=pl.lit(" y=x", dtype=pl.String), __point__=pl.lit(False, dtype=pl.Boolean))

    df_socres = [
        df.select(s, "__actual__")
        .with_columns(
            pl.col(s).qcut(n_bins, labels=[str(i) for i in range(n_bins)]).alias("__qcuts__")
        )
        .group_by("__qcuts__")
        .agg(
            mean_predicted_prob=pl.col(s).mean().cast(pl.Float64),
            fraction_of_positives=pl.col("__actual__").mean().cast(pl.Float64),
        )
        .sort("__qcuts__")
        .select(
            "mean_predicted_prob",
            "fraction_of_positives",
            score=pl.lit(s, dtype=pl.String),
            __point__=pl.lit(False, dtype=pl.Boolean),
        )
        for s in new_dict.keys()
        if s != "__actual__"
    ]

    chart1 = (
        alt.Chart(df_line)
        .mark_line(point=False)
        .encode(
            x="mean_predicted_prob:Q",
            y="fraction_of_positives:Q",
            color="score:N",
            strokeDash=alt.value([5, 5]),
        )
    )
    selection = alt.selection_multi(fields=["score"], bind="legend")
    chart2 = (
        alt.Chart(pl.concat(df_socres))
        .mark_line(point=True)
        .encode(
            x="mean_predicted_prob:Q",
            y="fraction_of_positives:Q",
            color="score:N",
            tooltip=["mean_predicted_prob", "fraction_of_positives"],
            opacity=alt.condition(selection, alt.value(0.8), alt.value(0.1)),
        )
        .add_params(selection)
    )

    return chart1 + chart2


def plot_roc_auc(
    *,
    target: Iterable[int],
    pred: pl.Series | None = None,
    name: str | None = None,
    preds: List[pl.Series] | None = None,
    names: List[str] | None = None,
    show_auc: bool = True,
    estimator_name: str = "",
    n_decimals: int = 4,
    **kwargs,
) -> alt.Chart:
    """
    Paramters
    ---------
    target
        A column which has the actual binary target information
    pred
        The prediction probability variable
    name
        The name for the prediction
    preds
        The prediction probability variables
    names
        The names for the predictions
    show_auc
        Whether to show the AUC value or not
    n_decimals
        Round to n-th decimal digit if show_auc is True
    kwargs
        Other keyword arguments to Altair's mark_line
    """

    if pred is not None:
        if name is None:
            raise ValueError("If `pred` is not None, then `name` must not be none.")
        else:
            new_dict = {name: pl.Series(values=pred)}

    else:  # pred is None
        if (preds is None) or (names is None):
            raise ValueError("If `pred` is None, then `preds` and `names` must be populated.")

        if hasattr(preds, "__len__") and (hasattr(names, "__len__")):
            if len(preds) != len(names):
                raise ValueError("Input `preds` and `names` must have the same length.")

            new_dict = {n: pl.Series(values=s) for n, s in zip(names, preds)}
        else:
            raise ValueError("Input `preds` and `names` must be iterables with a length.")

    target_series = pl.Series(name="__actual__", values=target)

    if any(len(s) != len(target_series) for s in new_dict.values()):
        raise ValueError("All input `score(s)` and `target` must have the same length.")

    pred_names = list(new_dict.keys())
    new_dict["__actual__"] = target_series
    df_tmp = pl.from_dict(new_dict)
    dfs = []

    for p in pred_names:
        zero = pl.DataFrame(
            {
                "tpr": [0.0],
                "fpr": [0.0],
            },
            schema={
                "tpr": pl.Float64,
                "fpr": pl.Float64,
            },
        )
        tpr_fpr = (
            df_tmp.select(tpr_fpr=query_tpr_fpr("__actual__", p).reverse())
            .unnest("tpr_fpr")
            .select("tpr", "fpr")
        )

        text = p
        if show_auc:
            auc = tpr_fpr.select(integrate_trapz("tpr", "fpr")).item(0, 0)
            text += f" (AUC = {round(auc, n_decimals)})"

        dfs.append(pl.concat([zero, tpr_fpr]).with_columns(name=pl.lit(text, dtype=pl.String)))

    perfect_line = pl.int_range(1, 100, step=5, eager=True) / 100
    df_line = pl.DataFrame({"fpr": perfect_line, "tpr": perfect_line}).with_columns(
        name=pl.lit(" y=x", dtype=pl.String),
    )
    chart1 = (
        alt.Chart(df_line)
        .mark_line()
        .encode(
            x=alt.X("fpr", title="False Positive Rate"),
            y=alt.Y("tpr", title="True Positive Rate"),
            color="name:N",
            strokeDash=alt.value([5, 5]),
        )
    )
    selection = alt.selection_multi(fields=["name"], bind="legend")
    chart2 = (
        alt.Chart(pl.concat(dfs))
        .mark_line(interpolate="step", **kwargs)
        .encode(
            x=alt.X("fpr", title="False Positive Rate"),
            y=alt.Y("tpr", title="True Positive Rate"),
            color="name:N",
            opacity=alt.condition(selection, alt.value(0.8), alt.value(0.1)),
        )
        .add_params(selection)
    )

    return chart1 + chart2



from __future__ import annotations
import importlib.metadata
import polars as pl
import polars_ds.config as Config  # noqa: F401

__version__ = importlib.metadata.version("polars_ds")

# Internal dependencies
from polars_ds.exprs import *  # noqa F403
from polars_ds.expander import E  # noqa: F401


def frame(size: int = 2_000, index_name: str = "row_num") -> pl.DataFrame:
    """
    Generates a frame with only an index (row number) column.
    This is a convenience function to be chained with pds.random(...) when running simulations and tests.

    Parameters
    ----------
    size
        The total number of rows in this dataframe
    index_name
        The name of the index column
    """
    return pl.DataFrame({index_name: range(size)})



from .expr_knn import *  # noqa : F403
from .expr_linear import *  # noqa : F403
from .metrics import *  # noqa : F403
from .num import *  # noqa : F403
from .stats import *  # noqa : F403
from .string import *  # noqa : F403
from .ts_features import *  # noqa : F403
from .expr_iter import *  # noqa : F403



from __future__ import annotations

import polars as pl
import json
from io import StringIO
from enum import Enum
from polars_ds._utils import to_expr
from polars_ds.typing import FitTransformFunc, ExprTransform
from typing import List, Sequence, Dict, Protocol, Literal
from polars._typing import IntoExprColumn


class PLContext(Enum):
    # Regular ExpreStep (purely expressed-based)
    SELECT = "select"
    WITH_COLUMNS = "with_columns"
    FILTER = "filter"
    EXPLODE = "explode"
    # SQLStep
    SQL = "sql"
    # SortStep
    SORT = "sort"
    # GROUP_BY_AGG
    GROUP_BY_AGG = "group_by_agg"
    # GROUP_BY_DYNAMIC_AGG
    GROUP_BY_DYN_AGG = "group_by_dyn_agg"

    def get_context_dict(self) -> Dict:
        return {"context": self.value}

    def build_step(self, deser_dict: Dict) -> PipelineStep:
        """
        Build a step from an already deserialized dict. The dict may contain values of type list[str],
        which requires further deserialization by polars from str to pl.Expr.
        """
        if self in (PLContext.SELECT, PLContext.WITH_COLUMNS, PLContext.FILTER, PLContext.EXPLODE):
            return ExprStep.from_partial_dict(deser_dict)
        elif self == PLContext.SORT:
            return SortStep.from_partial_dict(deser_dict)
        elif self == PLContext.GROUP_BY_AGG:
            return GroupByAggStep.from_partial_dict(deser_dict)
        elif self == PLContext.GROUP_BY_DYN_AGG:
            return GroupByDynAggStep.from_partial_dict(deser_dict)
        elif self == PLContext.SQL:
            return SQLStep.from_partial_dict(deser_dict)
        else:
            raise ValueError(f"Unknown PLContext: {self}")


class PipelineStep(Protocol):
    context: PLContext

    def apply_df(self, df: pl.LazyFrame | pl.DataFrame) -> pl.LazyFrame | pl.DataFrame: ...
    def to_json(self) -> str: ...

    @staticmethod
    def from_partial_dict(data: Dict) -> "PipelineStep":
        """
        See MakeStep.make's docstrings.
        """
        ...


class MakeStep:
    __slots__ = ()

    @staticmethod
    def make(deser_dict: Dict) -> PipelineStep:
        """
        Constructs a PipelineStep from a partially deserialized dictionary. The dictionary may be partially
        deserialized because some fields may actually be list[str], but it should be deserialized to list[pl.Expr]
        where the string elements will be further deserialized to Polars Expressions based on context.
        """
        ctx = deser_dict.get("context", None)
        if ctx is None:
            raise ValueError(
                "There is no 'context' key in the dictionary. Input is not conformant."
            )
        else:
            return PLContext(ctx).build_step(deser_dict)


class SQLStep:
    """
    Container for SQL transformation on a Polars dataframe.
    """

    __slots__ = ("sql_str", "context")

    def __init__(self, sql_str: str):
        self.sql_str: str = sql_str
        self.context: PLContext = PLContext.SQL

    @staticmethod
    def from_partial_dict(d: Dict) -> "SQLStep":
        return SQLStep(sql_str=d["sql_str"])

    def apply_df(self, df: pl.LazyFrame | pl.DataFrame) -> pl.LazyFrame | pl.DataFrame:
        return pl.SQLContext(df=df).execute(self.sql_str)

    def to_json(self) -> str:
        d = self.context.get_context_dict() | {"sql_str": self.sql_str}
        return json.dumps(d)


class SortStep:
    """
    Container for a polars df.sort() transform.
    """

    __slots__ = ("by", "descending", "context", "maintain_order")

    def __init__(
        self,
        by: str | pl.Expr | Sequence[str] | Sequence[pl.Expr],
        descending: bool | Sequence[bool],
        maintain_order: bool = True,
    ):
        if isinstance(by, (str, pl.Expr)):
            self.by: List[pl.Expr] = [to_expr(by)]
        elif isinstance(by, list):
            self.by: List[pl.Expr] = [to_expr(e) for e in by]
        else:
            raise ValueError("Input `by` can only be scalar str/pl.Expr or a list of them.")

        if isinstance(descending, bool):
            self.descending: List[bool] = [descending]
        elif isinstance(descending, list):
            self.descending: List[bool] = [bool(b) for b in descending]
        else:
            raise ValueError("Input `descending` must be a scalar bool or a list of bools.")

        if len(self.by) != len(self.descending):
            raise ValueError(
                f"Input `by` (len {len(by)}) doesn't match the length of `descending` (len {len(self.descending)})."
            )

        self.context: PLContext = PLContext.SORT
        self.maintain_order = maintain_order

    @staticmethod
    def from_partial_dict(d: Dict) -> "SortStep":
        by = [pl.Expr.deserialize(StringIO(e), format="json") for e in d["by"]]
        descending = [bool(e) for e in d["descending"]]
        maintain_order = bool(d["maintain_order"])
        return SortStep(by=by, descending=descending, maintain_order=maintain_order)

    def apply_df(self, df: pl.LazyFrame | pl.DataFrame) -> pl.LazyFrame | pl.DataFrame:
        return df.sort(by=self.by, descending=self.descending, maintain_order=self.maintain_order)

    def to_json(self) -> str:
        d = self.context.get_context_dict() | {
            "by": [e.meta.serialize(format="json") for e in self.by],
            "descending": self.descending,
            "maintain_order": self.maintain_order,
        }
        return json.dumps(d)


class GroupByAggStep:
    """
    Container for a polars df.group_by(...).agg(...) transform.
    """

    __slots__ = ("by", "agg", "maintain_order", "context")

    def __init__(
        self,
        by: str | pl.Expr | Sequence[str] | Sequence[pl.Expr],
        agg: List[pl.Expr],
        maintain_order: bool = False,
    ):
        if isinstance(by, (str, pl.Expr)):
            self.by: List[pl.Expr] = [to_expr(by)]
        elif isinstance(by, list):
            self.by: List[pl.Expr] = [to_expr(e) for e in by]
        else:
            raise ValueError("Input `by` can only be scalar str/pl.Expr or a list of them.")

        self.agg: List[pl.Expr] = agg
        self.context: PLContext = PLContext.GROUP_BY_AGG
        self.maintain_order = maintain_order

    @staticmethod
    def from_partial_dict(d: dict) -> "GroupByAggStep":
        by = [pl.Expr.deserialize(StringIO(e), format="json") for e in d["by"]]
        agg = [pl.Expr.deserialize(StringIO(e), format="json") for e in d["agg"]]
        return GroupByAggStep(by=by, agg=agg, maintain_order=d["maintain_order"])

    def apply_df(self, df: pl.LazyFrame | pl.DataFrame) -> pl.LazyFrame | pl.DataFrame:
        return df.group_by(self.by, maintain_order=self.maintain_order).agg(self.agg)

    def to_json(self) -> str:
        d = self.context.get_context_dict() | {
            "by": [e.meta.serialize(format="json") for e in self.by],
            "agg": [e.meta.serialize(format="json") for e in self.agg],
            "maintain_order": self.maintain_order,
        }
        return json.dumps(d)


class GroupByDynAggStep:
    """
    Container for a polars df.group_by_dynamic(...).agg(...) transform.
    """

    __slots__ = (
        "index_column",
        "every",
        "agg",
        "group_by",
        "period",
        "offset",
        "include_boundaries",
        "closed",
        "label",
        "start_by",
        "context",
    )

    def __init__(
        self,
        index_column: str,
        every: str,
        agg: List[pl.Expr],
        period: str | None = None,
        offset: str | None = None,
        include_boundaries: bool = False,
        closed: Literal["left", "right", "both", "none"] = "left",
        label: Literal["left", "right", "datapoint"] = "left",
        group_by: str | pl.Expr | Sequence[str] | Sequence[pl.Expr] | None = None,
        start_by: Literal[
            "window",
            "datapoint",
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        ] = "window",
    ):
        if closed not in ["left", "right", "both", "none"]:
            raise ValueError(
                "Input `closed` must be one of ['left', 'right', 'both', 'none']. See polars's group_by_dynamic for more info."
            )

        if label not in ["left", "right", "datapoint"]:
            raise ValueError(
                "Input `closed` must be one of ['left', 'right', 'datapoint']. See polars's group_by_dynamic for more info."
            )

        if start_by not in [
            "window",
            "datapoint",
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        ]:
            raise ValueError(
                "Input `closed` must be one of ['window', 'datapoint', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday']. See polars's group_by_dynamic for more info."
            )

        # Because of a strange bug.
        if isinstance(index_column, pl.Expr):
            raise ValueError(
                "Input `index_column` must be a string (name of a column) and not a Polars expression."
            )

        self.index_column: str = index_column
        self.every: str = every
        self.period: str | None = period
        self.offset: str | None = offset
        self.include_boundaries: bool = include_boundaries
        self.closed: Literal["left", "right", "both", "none"] = closed
        self.label: Literal["left", "right", "datapoint"] = label
        self.start_by: Literal[
            "window",
            "datapoint",
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        ] = start_by

        if isinstance(group_by, (str, pl.Expr)):
            self.group_by: List[pl.Expr] | None = [to_expr(group_by)]
        elif isinstance(group_by, list):
            self.group_by: List[pl.Expr] | None = [to_expr(e) for e in group_by]
        elif group_by is None:
            self.group_by: List[pl.Expr] | None = None
        else:
            raise ValueError(
                "Input `by` can only be scalar str / pl.Expr or a list of them or None."
            )

        self.agg: List[pl.Expr] = agg
        self.context: PLContext = PLContext.GROUP_BY_DYN_AGG

    @staticmethod
    def from_partial_dict(d: Dict) -> "GroupByDynAggStep":
        return GroupByDynAggStep(
            index_column=d["index_column"],
            every=d["every"],
            period=d["period"],
            offset=d["offset"],
            include_boundaries=d["include_boundaries"],
            label=d["label"],
            start_by=d["start_by"],
            group_by=None
            if d["group_by"] is None
            else [pl.Expr.deserialize(StringIO(e), format="json") for e in d["group_by"]],
            agg=[pl.Expr.deserialize(StringIO(e), format="json") for e in d["agg"]],
        )

    def apply_df(self, df: pl.LazyFrame | pl.DataFrame) -> pl.LazyFrame | pl.DataFrame:
        return df.group_by_dynamic(
            index_column=self.index_column,
            every=self.every,
            period=self.period,
            offset=self.offset,
            include_boundaries=self.include_boundaries,
            label=self.label,
            group_by=self.group_by,
            start_by=self.start_by,
        ).agg(self.agg)

    def to_json(self) -> str:
        d = self.context.get_context_dict() | {
            "group_by": None
            if self.group_by is None
            else [e.meta.serialize(format="json") for e in self.group_by],
            "agg": [e.meta.serialize(format="json") for e in self.agg],
            "index_column": self.index_column,
            "every": self.every,
            "period": self.period,
            "offset": self.offset,
            "include_boundaries": self.include_boundaries,
            "closed": self.closed,
            "label": self.label,
            "start_by": self.start_by,
        }
        return json.dumps(d)


class ExprStep:
    """
    Container for either one of these polars transforms:

    1. df.select(...)
    2. df.with_columns(...)
    3. df.filter(...)
    """

    __slots__ = ("exprs", "context")

    def __init__(self, exprs: pl.Expr | Sequence[pl.Expr], context: str | PLContext):
        self.context: PLContext = context if isinstance(context, PLContext) else PLContext(context)
        if isinstance(exprs, pl.Expr):
            self.exprs = [exprs]
        elif isinstance(exprs, str):
            self.exprs = [pl.col(exprs)]
        elif isinstance(exprs, list):
            self.exprs = [to_expr(e) for e in exprs]
        else:
            raise ValueError(
                "A pipeline step must be either an expression or a list of expressions."
            )

    @staticmethod
    def from_partial_dict(d: dict) -> "ExprStep":
        context, json_exprs = d["context"], d["exprs"]
        step_context = PLContext(context)
        if step_context in (PLContext.SELECT, PLContext.WITH_COLUMNS, PLContext.EXPLODE):
            exprs = [pl.Expr.deserialize(StringIO(e), format="json") for e in json_exprs]
        elif step_context == PLContext.FILTER:
            exprs = [pl.Expr.deserialize(StringIO(json_exprs), format="json")]
        elif step_context == PLContext.SQL:
            exprs = str(json_exprs)  # SQL context json_exprs is just a str
        else:
            raise ValueError("Input is not a valid PDS pipeline.")

        return ExprStep(exprs=exprs, context=step_context)

    def to_json(self) -> str:
        ctx = self.context
        d = ctx.get_context_dict()
        if ctx in (PLContext.SELECT, PLContext.WITH_COLUMNS, PLContext.EXPLODE):
            d["exprs"] = [e.meta.serialize(format="json") for e in self.exprs]
        elif self.context == PLContext.SQL:
            d["exprs"] = self.exprs[0]
        elif self.context == PLContext.FILTER:
            d["exprs"] = self.exprs[0].meta.serialize(format="json")
        else:  # Should never reach here
            raise ValueError(f"Unknown context: {self.context}")

        return json.dumps(d)

    def apply_df(self, df: pl.LazyFrame | pl.DataFrame) -> pl.LazyFrame | pl.DataFrame:
        if self.context == PLContext.SELECT:
            return df.select(self.exprs)
        elif self.context == PLContext.WITH_COLUMNS:
            return df.with_columns(self.exprs)
        elif self.context == PLContext.SQL:
            return pl.SQLContext(df=df, eager=False).execute(self.exprs[0])
        elif self.context == PLContext.FILTER:
            return df.filter(self.exprs[0])
        elif self.context == PLContext.EXPLODE:
            return df.explode(self.exprs)


class FitStep:
    # This doesn't satisfy the 'PipelineStep' protocol
    # This is turned into an ExprStep at the time of fit(materialization) in a blueprint.

    __slots__ = ("func", "cols", "exclude", "context")

    def __init__(
            self, 
            func: FitTransformFunc, 
            cols: IntoExprColumn | None, 
            exclude: List[str],
            context: PLContext = PLContext.WITH_COLUMNS
        ):
        self.func: FitTransformFunc = func
        self.cols: IntoExprColumn | None = cols
        self.exclude: List[str] = exclude
        self.context = context

    # Here we allow IntoExprColumn as input so that users can use selectors, or other polars expressions
    # to specify input columns, which adds flexibility.
    # We still need real column names so that the functions in transforms.py will work.
    def fit(self, df: pl.DataFrame | pl.LazyFrame) -> ExprTransform:
        if self.cols is None:
            return self.func(df)
        else:
            real_cols: List[str] = [
                x
                for x in df.lazy().select(self.cols).collect_schema().names()
                if x not in self.exclude
            ]
            return self.func(df, real_cols)


# ---------------------------------------------------------------------------------------
# Steps that don't need fitting, can be directly serialized.
# FitStep becomes ExprStep after fitting
_SERIALIZABLE_STEPS = [ExprStep, SortStep, SQLStep, GroupByAggStep, GroupByDynAggStep]



"""String-related utils, including edit distances, simple cleaning, etc."""

from __future__ import annotations

import warnings
import polars as pl
from typing import List, Literal, Dict

# Internal dependencies
from polars_ds._utils import pl_plugin, to_expr

__all__ = [
    "filter_by_levenshtein",
    "filter_by_hamming",
    "to_camel_case",
    "to_snake_case",
    "to_pascal_case",
    "to_constant_case",
    "str_nearest",
    "str_jaccard",
    "str_hamming",
    "str_sorensen_dice",
    "str_tversky_sim",
    "str_jw",
    "str_jaro",
    "str_d_leven",
    "str_leven",
    "str_osa",
    "str_lcs_seq",  # deprecated
    "str_lcs_subseq",
    "str_lcs_subseq_dist",
    "str_lcs_substr",
    "str_fuzz",
    "similar_to_vocab",
    "extract_numbers",
    "replace_non_ascii",
    "remove_diacritics",
    "normalize_string",
    "map_words",
    "normalize_whitespace",
]


def filter_by_levenshtein(
    c: str | pl.Expr,
    other: str | pl.Expr,
    bound: int,
    parallel: bool = False,
) -> pl.Expr:
    """
    Returns whether the Levenshtein distance between self and other is <= bound. This is
    faster than computing levenshtein distance and then doing a filter.

    Parameters
    ----------
    c
        Either the name of the column or a Polars expression
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    bound
        Closed upper bound. If distance <= bound, return true and false otherwise.
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """

    return pl_plugin(
        symbol="pl_levenshtein_filter",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(abs(bound), pl.UInt32),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def filter_by_hamming(
    c: str | pl.Expr,
    other: str | pl.Expr,
    bound: int,
    pad: bool = False,
    parallel: bool = False,
) -> pl.Expr:
    """
    Returns whether the hamming distance between self and other is <= bound. This is
    faster than computing hamming distance and then doing a filter. Note this does not pad
    the strings. If the lengths of the two strings do not match, they will be filtered out.

    Parameters
    ----------
    c
        Either the name of the column or a Polars expression
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    bound
        Closed upper bound. If distance <= bound, return true and false otherwise.
    parallel
        Whether to run the comparisons in parallel. Note that this is not always faster, especially
        when used with other expressions or in group_by/over context.
    """

    return pl_plugin(
        symbol="pl_hamming_filter",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(bound, dtype=pl.UInt32),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def str_hamming(
    c: str | pl.Expr, other: str | pl.Expr, pad: bool = False, parallel: bool = False
) -> pl.Expr:
    """
    Computes the hamming distance between two strings. If they do not have the same length, null will
    be returned.

    Parameters
    ----------
    c
        Either the name of the column or a Polars expression
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    pad
        Whether to pad the string when lengths are not equal.
    parallel
        Whether to run the comparisons in parallel. Note that this is not always faster, especially
        when used with other expressions or in group_by/over context.
    """

    if pad:
        return pl_plugin(
            symbol="pl_hamming_padded",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )
    else:
        return pl_plugin(
            symbol="pl_hamming",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )


def to_camel_case(c: str | pl.Expr) -> pl.Expr:
    """Turns itself into camel case. E.g. helloWorld"""
    return pl_plugin(
        symbol="pl_to_camel",
        args=[to_expr(c)],
        is_elementwise=True,
    )


def to_snake_case(c: str | pl.Expr) -> pl.Expr:
    """Turns itself into snake case. E.g. hello_world"""
    return pl_plugin(
        symbol="pl_to_snake",
        args=[to_expr(c)],
        is_elementwise=True,
    )


def to_pascal_case(c: str | pl.Expr) -> pl.Expr:
    """Turns itself into Pascal case. E.g. HelloWorld"""
    return pl_plugin(
        symbol="pl_to_pascal",
        args=[to_expr(c)],
        is_elementwise=True,
    )


def to_constant_case(c: str | pl.Expr) -> pl.Expr:
    """Turns itself into constant case. E.g. Hello_World"""
    return pl_plugin(
        symbol="pl_to_constant",
        args=[to_expr(c)],
        is_elementwise=True,
    )


def str_nearest(
    c: str | pl.Expr,
    word: str,
    threshold: int = 100,
    metric: Literal["lv", "hamming"] = "lv",
) -> pl.Expr:
    """
    Finds the string in the column that is nearest to the given word in the given metric. This algorithm is
    very slow.

    Note: Nearest-k strings search functionality is temporarily dropped.

    Parameters
    ----------
    c
        The string column or its name
    word
        Any iterable collection of strings that can be turned into a polars Series, or an expression
    threshold : int
        Only considers strings to be near if they are within distance threshold. This is a positive integer
        because all the distances output integers.
    metric
        Which similarity metric to use. One of `lv`, `hamming`
    """
    if metric not in ("lv", "hamming"):
        raise ValueError(f"Unknown metric for similar_words: {metric}")

    if threshold <= 0:
        raise ValueError("Distance threshold must be > 0.")

    return pl_plugin(
        symbol="pl_nearest_str",
        args=[to_expr(c)],
        kwargs={
            "word": word,
            "metric": str(metric).lower(),
            "threshold": threshold,
        },
        returns_scalar=True,
    )


def str_jaccard(
    c: str | pl.Expr,
    other: str | pl.Expr,
    substr_size: int = 2,
    parallel: bool = False,
) -> pl.Expr:
    """
    Treats substrings of size `substr_size` as a set. And computes the jaccard similarity between
    this word and the other.

    Note this treats substrings at the byte level under the hood, not at the char level. So non-ASCII
    characters may have problems.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    substr_size
        The substring size for Jaccard similarity. E.g. if substr_size = 2, "apple" will be decomposed into
        the set ('ap', 'pp', 'pl', 'le') before being compared.
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_str_jaccard",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(substr_size, pl.UInt32),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def str_overlap_coeff(
    c: str | pl.Expr,
    other: str | pl.Expr,
    substr_size: int = 2,
    parallel: bool = False,
) -> pl.Expr:
    """
    Treats substrings of size `substr_size` as a set. And computes the overlap coefficient as
    similarity between this word and the other.

    Note this treats substrings at the byte level under the hood, not at the char level. So non-ASCII
    characters may have problems.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    substr_size
        The substring size for Jaccard similarity. E.g. if substr_size = 2, "apple" will be decomposed into
        the set ('ap', 'pp', 'pl', 'le') before being compared.
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_overlap_coeff",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(substr_size, pl.UInt32),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def str_sorensen_dice(
    c: str | pl.Expr,
    other: str | pl.Expr,
    substr_size: int = 2,
    parallel: bool = False,
) -> pl.Expr:
    """
    Treats substrings of size `substr_size` as a set. And computes the Sorensen-Dice similarity between
    this word and the other.

    Note this treats substrings at the byte level under the hood, not at the char level. So non-ASCII
    characters may have problems.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    substr_size
        The substring size for Jaccard similarity. E.g. if substr_size = 2, "apple" will be decomposed into
        the set ('ap', 'pp', 'pl', 'le') before being compared.
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_sorensen_dice",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(substr_size, pl.UInt32),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def str_tversky_sim(
    c: str | pl.Expr,
    other: str | pl.Expr,
    alpha: float,
    beta: float,
    substr_size: int = 2,
    parallel: bool = False,
) -> pl.Expr:
    """
    Treats substrings of size `substr_size` as a set. And computes the tversky_sim similarity between
    this word and the other. See the reference for information on how Tversky similarity is related
    the other ngram based similarity.

    Note this treats substrings at the byte level under the hood, not at the char level. So non-ASCII
    characters may have problems. Also note that alpha and beta are supposed to be weighting factors,
    but this doesn't check whether they satisfy the definition of weights and has to be chosen at the
    discretion of the user.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    alpha
        The first weighting factor. See reference
    beta
        The second weighting factor. See reference
    substr_size
        The substring size for Jaccard similarity. E.g. if substr_size = 2, "apple" will be decomposed into
        the set ('ap', 'pp', 'pl', 'le') before being compared.
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.

    Reference
    ---------
    https://yassineelkhal.medium.com/the-complete-guide-to-string-similarity-algorithms-1290ad07c6b7
    """
    if alpha < 0 or beta < 0:
        raise ValueError("Input `alpha` and `beta` must be >= 0.")

    return pl_plugin(
        symbol="pl_tversky_sim",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(substr_size, pl.UInt32),
            pl.lit(alpha, pl.Float64),
            pl.lit(beta, pl.Float64),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def str_jw(
    c: str | pl.Expr,
    other: str | pl.Expr,
    weight: float = 0.1,
    parallel: bool = False,
) -> pl.Expr:
    """
    Computes the Jaro-Winkler similarity between this and the other str.
    Jaro-Winkler distance = 1 - Jaro-Winkler sim.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    weight
        Weight for prefix. A typical value is 0.1.
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_jw",
        args=[
            to_expr(c),
            to_expr(other),
            pl.lit(weight, pl.Float64),
            pl.lit(parallel, pl.Boolean),
        ],
    )


def str_jaro(c: str | pl.Expr, other: str | pl.Expr, parallel: bool = False) -> pl.Expr:
    """
    Computes the Jaro similarity between this and the other str. Jaro distance = 1 - Jaro sim.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_jaro",
        args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        is_elementwise=True,
    )


def str_d_leven(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
    return_sim: bool = False,
) -> pl.Expr:
    """
    Computes the Damerau-Levenshtein distance between this and the other str.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    return_sim
        If true, return normalized Damerau-Levenshtein.
    """
    if return_sim:
        return pl_plugin(
            symbol="pl_d_levenshtein_sim",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )
    else:
        return pl_plugin(
            symbol="pl_d_levenshtein",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )


def str_leven(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
    return_sim: bool = False,
) -> pl.Expr:
    """
    Computes the Levenshtein distance between this and the other str.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    return_sim
        If true, return normalized Levenshtein.
    """
    if return_sim:
        return pl_plugin(
            symbol="pl_levenshtein_sim",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )
    else:
        return pl_plugin(
            symbol="pl_levenshtein",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )


def str_lcs_substr(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
) -> pl.Expr:
    """
    Extracts the longest common substring from the string between this and the other string.

    Note: this is not the same as the longest common subsequence.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_lcs_substr",
        args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
    )


def str_lcs_subseq(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
) -> pl.Expr:
    """
    Extracts the longest common subsequence from the string between this and the other string.

    Note: this is not the same as the longest common substring.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_lcs_subseq",
        args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
    )


def str_lcs_seq(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
    return_sim: bool = True,
) -> pl.Expr:
    warnings.warn(
        "`str_lcs_seq` is deprecated and users should use `str_lcs_subseq_dist`.",
        DeprecationWarning,
        stacklevel=2,
    )
    return str_lcs_subseq_dist(c, other, parallel, return_sim)


def str_lcs_subseq_dist(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
    return_sim: bool = True,
) -> pl.Expr:
    """
    Computes the Longest Common Subsequence distance/similarity between this and the other str.
    The distance is calculated as max(len1, len2) - similarity, where the similarity is the
    the length of the longest common subsequence. Subsequences may not occupy consecutive positions.

    The subsequence does not need to be consecutive.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    return_sim
        If true, return normalized similarity.
    """
    if return_sim:
        return pl_plugin(
            symbol="pl_lcs_subseq_sim",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )
    else:
        return pl_plugin(
            symbol="pl_lcs_subseq_dist",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )


def str_osa(
    c: str | pl.Expr,
    other: str | pl.Expr,
    parallel: bool = False,
    return_sim: bool = False,
) -> pl.Expr:
    """
    Computes the Optimal String Alignment distance between this and the other str.

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    return_sim
        If true, return normalized OSA similarity.
    """
    if return_sim:
        return pl_plugin(
            symbol="pl_osa_sim",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )
    else:
        return pl_plugin(
            symbol="pl_osa",
            args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
        )


def str_fuzz(c: str | pl.Expr, other: str | pl.Expr, parallel: bool = False) -> pl.Expr:
    """
    Calculates the normalized Indel similarity. (See the package rapidfuzz, fuzz.ratio for more
    information.)

    Parameters
    ----------
    c
        The string column
    other
        Either the name of the column or a Polars expression. If you want to compare a single
        string with all of column c, use pl.lit(your_str)
    parallel
        Whether to run the comparisons in parallel. Note that this is only recommended when this query
        is the only one in the context and we are not in any aggregation context.
    """
    return pl_plugin(
        symbol="pl_fuzz",
        args=[to_expr(c), to_expr(other), pl.lit(parallel, pl.Boolean)],
    )


def similar_to_vocab(
    c: str | pl.Expr,
    vocab: List[str],
    threshold: float,
    metric: Literal["lv", "dlv", "jw", "osa"] = "lv",
    strategy: Literal["avg", "all", "any"] = "avg",
) -> pl.Expr:
    """
    Compare each word in the vocab with each word in the column c. Returns a boolean
    that indicates whether there exist words in c that are similar to words in vocab.

    Parameters
    ----------
    c
        The string column
    vocab
        Any iterable collection of strings
    threshold
        A entry is considered similar to the words in the vocabulary if the similarity
        is above (>=) the threshold
    metric
        Which similarity metric to use. One of `lv`, `dlv`, `jw`, `osa`
    strategy
        If `avg`, then will return true if the average similarity is above the threshold.
        If `all`, then will return true if the similarity to all words in the vocab is above
        the threshold.
        If `any`, then will return true if the similarity to any words in the vocab is above
        the threshold.
    """
    if metric == "lv":
        sims = [str_leven(c, pl.lit(w, dtype=pl.String), return_sim=True) for w in vocab]
    elif metric == "dlv":
        sims = [str_d_leven(c, pl.lit(w, dtype=pl.String), return_sim=True) for w in vocab]
    elif metric == "osa":
        sims = [str_osa(c, pl.lit(w, dtype=pl.String), return_sim=True) for w in vocab]
    elif metric == "jw":
        sims = [str_jw(c, pl.lit(w, dtype=pl.String), return_sim=True) for w in vocab]
    else:
        raise ValueError(f"Unknown metric: {metric}")

    if strategy == "all":
        return pl.all_horizontal(s >= threshold for s in sims)
    elif strategy == "any":
        return pl.any_horizontal(s >= threshold for s in sims)
    elif strategy == "avg":
        return (pl.sum_horizontal(sims) / len(vocab)) >= threshold
    else:
        raise ValueError(f"Unknown strategy: {strategy}")


def extract_numbers(
    c: str | pl.Expr,
    ignore_comma: bool = False,
    join_by: str = "",
    dtype: pl.DataType = pl.String,
) -> pl.Expr:
    """
    Extracts numbers from the string column, and stores them in a list.

    Parameters
    ----------
    c
        The string column
    ignore_comma
        Whether to remove all comma before matching for numbers
    join_by
        If dtype is pl.String, join the list of strings using the value given here
    dtype
        The desired inner dtype for the extracted data. Should either be one of
        one of Polars' numerical types or pl.String

    Examples
    --------
    >>> df = pl.DataFrame(
    ...     {
    ...         "survey": [
    ...             "0% of my time",
    ...             "1% to 25% of my time",
    ...             "75% to 99% of my time",
    ...             "50% to 74% of my time",
    ...             "75% to 99% of my time",
    ...             "50% to 74% of my time",
    ...         ]
    ...     }
    ... )
    >>> df.select(pl.col("survey").str_ext.extract_numbers(dtype=pl.UInt32))
    shape: (6, 1)
    ┌───────────┐
    │ survey    │
    │ ---       │
    │ list[u32] │
    ╞═══════════╡
    │ [0]       │
    │ [1, 25]   │
    │ [75, 99]  │
    │ [50, 74]  │
    │ [75, 99]  │
    │ [50, 74]  │
    └───────────┘
    >>> df.select(pl.col("survey").str_ext.extract_numbers(join_by="-", dtype=pl.String))
    shape: (6, 1)
    ┌────────┐
    │ survey │
    │ ---    │
    │ str    │
    ╞════════╡
    │ 0      │
    │ 1-25   │
    │ 75-99  │
    │ 50-74  │
    │ 75-99  │
    │ 50-74  │
    └────────┘
    """
    expr = to_expr(c)
    if ignore_comma:
        expr = expr.str.replace_all(",", "")

    # Find all numbers
    expr = expr.str.extract_all(r"(\d*\.?\d+)")

    if dtype in [
        pl.UInt8,
        pl.UInt16,
        pl.UInt32,
        pl.UInt64,
        pl.Int8,
        pl.Int16,
        pl.Int32,
        pl.Int64,
        pl.Float32,
        pl.Float64,
    ]:
        expr = expr.list.eval(pl.element().cast(dtype))
    elif dtype == pl.String:  # As a list of strings
        if join_by != "":
            expr = expr.list.join(join_by)

    return expr


def replace_non_ascii(c: str | pl.Expr, value: str = "") -> pl.Expr:
    """Replaces non-Ascii values with the specified value.

    Parameters
    ----------
    c : str | pl.Expr
        The column name or expression
    value : str
        The value to replace non-Ascii values with, by default ""

    Returns
    -------
    pl.Expr

    Examples
    --------
    >>> df = pl.DataFrame({"x": ["mercy", "xbĤ", "ĤŇƏ"]})
    >>> df.select(pds.replace_non_ascii("x"))
    shape: (3, 1)
    ┌───────┐
    │ x     │
    │ ---   │
    │ str   │
    ╞═══════╡
    │ mercy │
    │ xb    │
    │       │
    └───────┘
    """
    expr = to_expr(c)

    if value == "":
        return pl_plugin(
            symbol="remove_non_ascii",
            args=[expr],
            is_elementwise=True,
        )

    return expr.str.replace_all(r"[^\p{Ascii}]", value)


def remove_diacritics(c: str | pl.Expr) -> pl.Expr:
    """Remove diacritics (e.g. è -> e) by converting the string to its NFD normalized
    form and removing the resulting non-ASCII components.

    Parameters
    ----------
    c : str | pl.Expr

    Returns
    -------
    pl.Expr

    Examples
    --------
    >>> df = pl.DataFrame({"x": ["mercy", "mèrcy"]})
    >>> df.select(pds.replace_non_ascii("x"))
    shape: (2, 1)
    ┌───────┐
    │ x     │
    │ ---   │
    │ str   │
    ╞═══════╡
    │ mercy │
    │ mercy │
    └───────┘
    """
    return pl_plugin(
        symbol="remove_diacritics",
        args=[to_expr(c)],
        is_elementwise=True,
    )


def normalize_string(c: str | pl.Expr, form: Literal["NFC", "NFKC", "NFD", "NFKD"]) -> pl.Expr:
    """
    Normalize Unicode string using one of 'NFC', 'NFKC', 'NFD', or 'NFKD'
    normalization.

    See https://en.wikipedia.org/wiki/Unicode_equivalence for more information.

    Parameters
    ----------
    c : str | pl.Expr
        The string column
    form: Literal["NFC", "NFKC", "NFD", "NFKD"]
        The Unicode normalization form to use

    Returns
    -------
    pl.Expr

    Examples
    --------
    >>> df = pl.DataFrame({"x": ["\u0043\u0327"], "y": ["\u00c7"]})
    >>> df.with_columns(
    >>>     pl.col("x").eq(pl.col("y")).alias("is_equal"),
    >>>     pds.normalize_string("x", "NFC")
    >>>     .eq(pds.normalize_string("y", "NFC"))
    >>>     .alias("normalized_is_equal"),
    >>> )
    shape: (1, 4)
    ┌─────┬─────┬──────────┬─────────────────────┐
    │ x   ┆ y   ┆ is_equal ┆ normalized_is_equal │
    │ --- ┆ --- ┆ ---      ┆ ---                 │
    │ str ┆ str ┆ bool     ┆ bool                │
    ╞═════╪═════╪══════════╪═════════════════════╡
    │ Ç   ┆ Ç   ┆ false    ┆ true                │
    └─────┴─────┴──────────┴─────────────────────┘
    """

    warnings.warn(
        "This function is deprecated because Polars>=1.20 has a native str.normalize() method. "
        + "This will be removed in a future version.",
        DeprecationWarning,
        stacklevel=2,
    )

    if form not in ("NFC", "NFKC", "NFD", "NFKD"):
        raise ValueError(
            f"{form} is not a valid Unicode normalization form.",
            " Please specify one of `NFC, NFKC, NFD, NFKD`",
        )

    return pl_plugin(
        symbol="normalize_string",
        args=[to_expr(c)],
        kwargs={"form": form},
        is_elementwise=True,
    )


def map_words(c: str | pl.Expr, mapping: Dict[str, str]) -> pl.Expr:
    """
    Replace words based on the specified mapping.

    Parameters
    ----------
    c : str | pl.Expr
        The string column
    mapping : dict[str, str]
        A dictionary of {word: the replacement}

    Returns
    -------
    pl.Expr

    Examples
    --------
    >>> df = pl.DataFrame({"x": ["one two three"]})
    >>> df.select(pds.map_words("x", {"two": "2"}))
    shape: (1, 1)
    ┌─────────────┐
    │ x           │
    │ ---         │
    │ str         │
    ╞═════════════╡
    │ one 2 three │
    └─────────────┘
    """
    return pl_plugin(
        symbol="map_words",
        args=[to_expr(c)],
        kwargs={"mapping": mapping},
        is_elementwise=True,
    )


def normalize_whitespace(c: str | pl.Expr, only_spaces: bool = False) -> pl.Expr:
    """
    Normalize whitespace to one, e.g. 'a   b' -> 'a b'.

    Parameters
    ----------
    c : str | pl.Expr
        The string column
    only_spaces: bool
        If True, only split on the space character ' ' instead of any whitespace
        character such as '\t' and '\n', by default False

    Returns
    -------
    pl.Expr

    Examples
    --------
    shape: (2, 3)
    ┌─────────┬─────┬────────┐
    │ x       ┆ y   ┆ z      │
    │ ---     ┆ --- ┆ ---    │
    │ str     ┆ str ┆ str    │
    ╞═════════╪═════╪════════╡
    │ a     b ┆ a b ┆ a b    │
    │ a	    b ┆ a b ┆ a	    b│
    └─────────┴─────┴────────┘
    """
    expr = to_expr(c)

    if only_spaces:
        return expr.str.replace_all(" +", " ")

    return pl_plugin(
        symbol="normalize_whitespace",
        args=[expr],
        is_elementwise=True,
    )



import polars as pl
import polars_ds as pds
import pytest
import scipy.stats
import sklearn.metrics

SEED = 208

SIZE = 1_000_000
DF = (
    pds.random_data(size=SIZE, n_cols=2)
    .rename({"feature_1": "y_true_score", "feature_2": "y_score"})
    .with_columns(
        pl.col("y_true_score").ge(0.5).alias("y_true"),
    )
    .drop("row_num")
)

MUTLICLASS_DF = (
    pds.random_data(size=SIZE, n_cols=2)
    .with_columns(
        pds.random_int(0, 2).alias("y_true"),
        pl.concat_list("feature_1", "feature_2").alias("y_score"),
    )
    .drop("feature_1", "feature_2")
)

SIZES = [1_000, 10_000, 100_000, 1_000_000]


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="mad")
def test_mad(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_mad("y_score"))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="mad")
def test_scipy_mad(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        scipy.stats.median_abs_deviation(df["y_score"])


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="r2")
def test_r2(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_r2("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="r2")
def test_sklearn_r2(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        sklearn.metrics.r2_score(df["y_true_score"], df["y_score"])


@pytest.mark.parametrize("n", SIZES)
def test_log_cosh(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_log_cosh("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_hubor_loss(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_hubor_loss("y_true_score", "y_score", delta=0.2))


@pytest.mark.parametrize("n", SIZES)
def test_l2(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_l2("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_l1(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_l1("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_l_inf(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_l_inf("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_log_loss(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_log_loss("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_mape(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_mape("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_smape(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_smape("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_msle(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_msle("y_true_score", "y_score"))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="roc_auc")
def test_roc_auc(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_roc_auc("y_true", "y_score"))


@pytest.mark.parametrize("n", SIZES)
@pytest.mark.benchmark(group="roc_auc")
def test_sklearn_roc_auc(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        sklearn.metrics.roc_auc_score(df["y_true"], df["y_score"])


@pytest.mark.parametrize("n", SIZES)
def test_binary_metrics(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_binary_metrics("y_true", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_confusion_matrix(benchmark, n):
    df = DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_confusion_matrix("y_true", "y_score"))


@pytest.mark.parametrize("n", SIZES)
def test_multi_roc_auc(benchmark, n):
    df = MUTLICLASS_DF.sample(n=n, seed=SEED)

    @benchmark
    def func():
        df.select(pds.query_multi_roc_auc("y_true", "y_score", n_classes=3))



"""
KNN related query expressions in Polars.
"""

from __future__ import annotations
import polars as pl
from typing import Any, Iterable, List, Sequence, cast
import warnings

# Internal dependencies
from polars_ds._utils import pl_plugin, to_expr
from polars_ds.typing import Distance

__all__ = [
    "query_knn_ptwise",
    "query_knn_freq_cnt",
    "query_knn_avg",
    "query_radius_ptwise",
    "query_radius_freq_cnt",
    "query_nb_cnt",
    "query_dist_from_kth_nb",
    "is_knn_from",
    "within_dist_from",
]


def warn_len_compare(item1: Iterable[Any], item2: Iterable[Any]) -> bool:
    """
    Compares the len of two Iterables if they have len returning true and warning if no len.

    Parameters
    ----------
    item1: Iterable[Any]
        Any iterable
    item2: Iterable[Any])
        Any iterable

    Returns:
        bool: If both items have __len__ then it will simply return whether or not
            they have equal size. If they don't have len then it returns True with a
            warning
    """
    # print()
    if hasattr(item1, "__len__") and hasattr(item2, "__len__"):
        return len(cast(Sequence, item1)) == len(cast(Sequence, item2))
    else:
        msg = "The inputs do not each have len so can't be compared, unexpected results may follow."
        warnings.warn(msg, stacklevel=2)
        return True


def query_dist_from_kth_nb(
    *features: str | pl.Expr,
    k: int,
    dist: Distance = "sql2",
    parallel: bool = False,
    epsilon: float = 0.0,
    max_bound: float = 99999.0,
) -> pl.Expr:
    """
    Computes the distance of each row to its k-th closest neighbor. This is useful for outlier detection.
    E.g. if the average distance to the 5th neighbor is 0.1, then a distance of 0.3 to the 5th neighbor might
    indicate that the point might be far away from neighboring points, or that it occupies a sparse region in which
    sample points typically do not appear.

    This can be 10% faster and more direct than getting the result from `query_knn_ptwise` with return_distance = True.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    k : int
        Number of neighbors to query
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    parallel : bool
        Whether to run the k-nearest neighbor query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    epsilon
        If > 0, then it is possible to miss a neighbor within epsilon distance away. This parameter
        should increase as the dimension of the vector space increases because higher dimensions
        allow for errors from more directions.
    max_bound
        Max distance the neighbors must be within
    """
    return pl_plugin(
        symbol="pl_dist_from_kth_nb",
        args=[to_expr(e) for e in features],
        kwargs={
            "k": k,
            "metric": str(dist).lower(),
            "parallel": parallel,
            "skip_eval": False,
            "max_bound": max_bound,
            "epsilon": epsilon,
        },
    )


def query_knn_ptwise(
    *features: str | pl.Expr,
    index: str | pl.Expr,
    k: int,
    dist: Distance = "sql2",
    parallel: bool = False,
    return_dist: bool = False,
    eval_mask: str | pl.Expr | None = None,
    data_mask: str | pl.Expr | None = None,
    epsilon: float = 0.0,
    max_bound: float = 99999.0,
) -> pl.Expr:
    """
    Takes the index column, and uses feature columns to determine the k nearest neighbors
    to each row. By default, this will return k + 1 neighbors, because the point (the row) itself
    is a neighbor to itself and this returns k additional neighbors. The only exception to this
    is when data_mask excludes the point from being a neighbor, in which case, k + 1 distinct neighbors will
    be returned. Any row with a null/NaN will never be a neighbor and will have null as its neighbor.

    Note that the index column must be convertible to u32. If you do not have a u32 column,
    you can generate one using pl.int_range(..), which should be a step before this. The index column
    must not contain nulls.

    Note that a default max distance bound of 99999.0 is applied. This means that if we cannot find
    k neighbors within `max_bound`, then there will be < k neighbors returned.

    Also note that this internally builds a kd-tree for fast querying and deallocates it once we
    are done. If you need to repeatedly run the same query on the same data, then it is not
    ideal to use this. A specialized external kd-tree structure would be better in that case.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    index : str | pl.Expr
        The column used as index, must be castable to u32
    k : int
        Number of neighbors to query
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    parallel : bool
        Whether to run the k-nearest neighbor query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    return_dist
        If true, return a struct with indices and distances.
    eval_mask
        Either None or a boolean expression or the name of a boolean column. If not none, this will
        only evaluate KNN for rows where this is true. This can speed up computation when only results on a
        subset are nedded.
    data_mask
        Either None or a boolean expression or the name of a boolean column. If none, all rows can be
        neighbors. If not None, the pool of possible neighbors will be rows where this is true.
    epsilon
        If > 0, then it is possible to miss a neighbor within epsilon distance away. This parameter
        should increase as the dimension of the vector space increases because higher dimensions
        allow for errors from more directions.
    max_bound
        Max distance the neighbors must be within
    """
    if k < 1:
        raise ValueError("Input `k` must be >= 1.")

    if dist in ("cosine", "h", "haversine"):
        raise ValueError(f"Distance {dist} doesn't work with current implementation.")

    idx = to_expr(index).cast(pl.UInt32).rechunk()
    cols = [idx]
    feats: List[pl.Expr] = [to_expr(e) for e in features]

    skip_data = data_mask is not None
    if skip_data:  # true means keep
        keep_mask = pl.all_horizontal(to_expr(data_mask), *(f.is_not_null() for f in feats))
    else:
        keep_mask = pl.all_horizontal(f.is_not_null() for f in feats)

    cols.append(keep_mask)
    skip_eval = eval_mask is not None
    if skip_eval:
        cols.append(to_expr(eval_mask))

    cols.extend(feats)
    kwargs = {
        "k": k,
        "metric": str(dist).lower(),
        "parallel": parallel,
        "skip_eval": skip_eval,
        "max_bound": max_bound,
        "epsilon": 0.0,
    }
    if return_dist:
        return pl_plugin(
            symbol="pl_knn_ptwise_w_dist",
            args=cols,
            kwargs=kwargs,
        )
    else:
        return pl_plugin(
            symbol="pl_knn_ptwise",
            args=cols,
            kwargs=kwargs,
        )


def query_knn_freq_cnt(
    *features: str | pl.Expr,
    index: str | pl.Expr,
    k: int,
    dist: Distance = "sql2",
    parallel: bool = False,
    eval_mask: str | pl.Expr | None = None,
    data_mask: str | pl.Expr | None = None,
    epsilon: float = 0.0,
    max_bound: float = 99999.0,
) -> pl.Expr:
    """
    Takes the index column, and uses feature columns to determine the k nearest neighbors
    to each row, and finally returns the number of times a row is a KNN of some other point.

    This calls `query_knn_ptwise` internally. See the docstring of `query_knn_ptwise` for more info.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    index : str | pl.Expr
        The column used as index, must be castable to u32
    k : int
        Number of neighbors to query
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    parallel : bool
        Whether to run the k-nearest neighbor query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    return_dist
        If true, return a struct with indices and distances.
    eval_mask
        Either None or a boolean expression or the name of a boolean column. If not none, this will
        only evaluate KNN for rows where this is true. This can speed up computation when only results on a
        subset are nedded.
    data_mask
        Either None or a boolean expression or the name of a boolean column. If none, all rows can be
        neighbors. If not None, the pool of possible neighbors will be rows where this is true.
    epsilon
        If > 0, then it is possible to miss a neighbor within epsilon distance away. This parameter
        should increase as the dimension of the vector space increases because higher dimensions
        allow for errors from more directions.
    max_bound
        Max distance the neighbors must be within
    """

    knn_expr: pl.Expr = query_knn_ptwise(
        *features,
        index=index,
        k=k,
        dist=dist,
        parallel=parallel,
        return_dist=False,
        eval_mask=eval_mask,
        data_mask=data_mask,
        epsilon=epsilon,
        max_bound=max_bound,
    )
    return knn_expr.explode().drop_nulls().value_counts(sort=True, parallel=parallel)


def query_knn_avg(
    *features: str | pl.Expr,
    target: str | pl.Expr,
    k: int,
    dist: Distance = "sql2",
    weighted: bool = False,
    parallel: bool = False,
    min_bound: float = 1e-9,
    max_bound: float = 99999.0,
) -> pl.Expr:
    """
    Takes the target column, and uses feature columns to determine the k nearest neighbors
    to each row. By default, this will return k + 1 neighbors, because the point (the row) itself
    is a neighbor to itself and this returns k additional neighbors. Any row with a null/NaN will
    never be a neighbor and will get null as the average.

    Note that a default max distance bound of 99999.0 is applied. This means that if we cannot find
    k neighbors within `max_bound`, then there will be < k neighbors returned.

    This is also known as KNN Regression, but really it is just the average of the K nearest neighbors.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    target : str | pl.Expr
        Float, must be castable to f64. This should not contain null.
    k : int
        Number of neighbors to query
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    weighted : bool
        If weighted, it will use 1/distance as weights to compute the KNN average. If min_bound is
        an extremely small value, this will default to 1/(1+distance) as weights to avoid division by 0.
    parallel : bool
        Whether to run the k-nearest neighbor query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    min_bound
        Min distance (>=) for a neighbor to be part of the average calculation. This prevents "identical"
        points from being part of the average and prevents division by 0. Note that this filter is applied
        after getting k nearest neighbors.
    max_bound
        Max distance the neighbors must be within (<)
    """
    if k < 1:
        raise ValueError("Input `k` must be >= 1.")

    if dist in ("cosine", "h", "haversine"):
        raise ValueError(f"Distance {dist} doesn't work with current implementation.")

    idx = to_expr(target).cast(pl.Float64).rechunk()
    feats = [to_expr(f) for f in features]
    keep_data = ~pl.any_horizontal(f.is_null() for f in feats)
    cols = [idx, keep_data]
    cols.extend(feats)

    kwargs = {
        "k": k,
        "metric": str(dist).lower(),
        "weighted": weighted,
        "parallel": parallel,
        "min_bound": min_bound,
        "max_bound": max_bound,
    }

    return pl_plugin(
        symbol="pl_knn_avg",
        args=cols,
        kwargs=kwargs,
    )


def within_dist_from(
    *features: str | pl.Expr,
    pt: Sequence[float] | Iterable[float],
    r: float | pl.Expr,
    dist: Distance = "sql2",
) -> pl.Expr:
    """
    Returns a boolean column that returns points that are within radius from the given point.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    pt : Iterable[float]
        The point
    r : either a float or an expression
        The radius to query with. If this is an expression, the radius will be applied row-wise.
    dist : Literal[`l1`, `l2`, `sql2`, `inf`, `cosine`, `haversine`]
        Note `sql2` stands for squared l2.
    """
    # For a single point, it is faster to just do it in native polars
    oth = [to_expr(x) for x in features]
    if not warn_len_compare(pt, oth):
        raise ValueError("Dimension does not match.")

    if dist == "l1":
        return (
            pl.sum_horizontal((e - pl.lit(xi, dtype=pl.Float64)).abs() for xi, e in zip(pt, oth))
            <= r
        )
    elif dist in ("l2", "sql2"):
        return (
            pl.sum_horizontal((e - pl.lit(xi, dtype=pl.Float64)).pow(2) for xi, e in zip(pt, oth))
            <= r
        )
    elif dist == "inf":
        return (
            pl.max_horizontal((e - pl.lit(xi, dtype=pl.Float64)).abs() for xi, e in zip(pt, oth))
            <= r
        )
    elif dist == "cosine":
        x_list = list(pt)
        x_norm = sum(z * z for z in x_list)
        oth_norm = pl.sum_horizontal(e * e for e in oth)
        distN = (
            1.0
            - pl.sum_horizontal(xi * e for xi, e in zip(x_list, oth)) / (x_norm * oth_norm).sqrt()
        )
        return distN <= r
    elif dist in ("h", "haversine"):
        from . import haversine

        pt_as_list = list(pt)
        if (len(pt_as_list) != 2) or (len(oth) < 2):
            raise ValueError(
                "For Haversine distance, input x must have dimension 2 and 2 other columns"
                " must be provided as lat and long."
            )

        y_lat = pl.lit(pt_as_list[0], dtype=pl.Float64)
        y_long = pl.lit(pt_as_list[1], dtype=pl.Float64)
        dist_out = haversine(oth[0], oth[1], y_lat, y_long)
        return dist_out <= r
    else:
        raise ValueError(f"Unknown distance function: {dist}")


def is_knn_from(
    *features: str | pl.Expr,
    pt: Iterable[float],
    k: int,
    dist: Distance = "sql2",
) -> pl.Expr:
    """
    Returns a boolean column that returns points that are k nearest neighbors from the point.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    pt : Iterable[float]
        The point
    k : int
        k nearest neighbor
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    """
    # For a single point, it is faster to just do it in native polars
    oth = [to_expr(x) for x in features]
    if not warn_len_compare(pt, oth):
        raise ValueError("Dimension does not match.")

    if dist == "l1":
        dist_out = pl.sum_horizontal(
            (e - pl.lit(xi, dtype=pl.Float64)).abs() for xi, e in zip(pt, oth)
        )
        return dist_out <= dist_out.bottom_k(k=k).max()
    elif dist in ("l2", "sql2"):
        dist_out = pl.sum_horizontal(
            (e - pl.lit(xi, dtype=pl.Float64)).pow(2) for xi, e in zip(pt, oth)
        )
        return dist_out <= dist_out.bottom_k(k=k).max()
    elif dist == "inf":
        dist_out = pl.max_horizontal(
            (e - pl.lit(xi, dtype=pl.Float64)).abs() for xi, e in zip(pt, oth)
        )
        return dist_out <= dist_out.bottom_k(k=k).max()
    elif dist == "cosine":
        x_list = list(pt)
        x_norm = sum(z * z for z in x_list)
        oth_norm = pl.sum_horizontal(e * e for e in oth)
        dist_out = (
            1.0
            - pl.sum_horizontal(xi * e for xi, e in zip(x_list, oth)) / (x_norm * oth_norm).sqrt()
        )
        return dist_out <= dist_out.bottom_k(k=k).max()
    elif dist in ("h", "haversine"):
        from . import haversine

        pt_as_list = list(pt)
        if (len(pt_as_list) != 2) or (len(oth) < 2):
            raise ValueError(
                "For Haversine distance, input x must have dimension 2 and 2 other columns"
                " must be provided as lat and long."
            )

        y_lat = pl.lit(pt_as_list[0], dtype=pl.Float64)
        y_long = pl.lit(pt_as_list[1], dtype=pl.Float64)
        dist_out = haversine(oth[0], oth[1], y_lat, y_long)
        return dist_out <= dist_out.bottom_k(k=k).max()
    else:
        raise ValueError(f"Unknown distance function: {dist}")


def query_radius_ptwise(
    *features: str | pl.Expr,
    index: str | pl.Expr,
    r: float,
    dist: Distance = "sql2",
    sort: bool = True,
    parallel: bool = False,
) -> pl.Expr:
    """
    Takes the index column, and uses features columns to determine distance, and finds all neighbors
    within distance r from each id. If you only care about neighbor count, you should use
    `query_nb_cnt`, which supports expression for radius and is way faster.

    Note that the index column must be convertible to u32. If you do not have a u32 ID column,
    you can generate one using pl.int_range(..), which should be a step before this.

    Also note that this internally builds a kd-tree for fast querying and deallocates it once we
    are done. If you need to repeatedly run the same query on the same data, then it is not
    ideal to use this. A specialized external kd-tree structure would be better in that case.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    index : str | pl.Expr
        The column used as index, must be castable to u32
    r : float
        The radius. Must be a scalar value now.
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    sort
        Whether the neighbors returned should be sorted by the distance. Setting this to False can
        improve performance by 10-20%.
    parallel : bool
        Whether to run the k-nearest neighbor query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    """

    if r <= 0.0:
        raise ValueError("Input `r` must be > 0.")
    elif isinstance(r, pl.Expr):
        raise ValueError("Input `r` must be a scalar now. Expression input is not implemented.")

    if dist in ("cosine", "h", "haversine"):
        raise ValueError(f"Distance {dist} doesn't work with current implementation.")

    idx = to_expr(index).cast(pl.UInt32).rechunk()
    metric = str(dist).lower()
    cols = [idx]
    cols.extend(to_expr(x) for x in features)
    return pl_plugin(
        symbol="pl_query_radius_ptwise",
        args=cols,
        kwargs={"r": r, "metric": metric, "parallel": parallel, "sort": sort},
    )


def query_radius_freq_cnt(
    *features: str | pl.Expr,
    index: str | pl.Expr,
    r: float,
    dist: Distance = "sql2",
    parallel: bool = False,
) -> pl.Expr:
    """
    Takes the index column, and uses features columns to determine distance, finds all neighbors
    within distance r from each index, and finally finds the count of the number of times the point is
    within distance r from other points.

    This calls `query_radius_ptwise` internally. See the docstring of `query_radius_ptwise` for more info.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    index : str | pl.Expr
        The column used as index, must be castable to u32
    r : float
        The radius. Must be a scalar value now.
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    parallel : bool
        Whether to run the k-nearest neighbor query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    """
    within_radius = query_radius_ptwise(
        *features, index=index, r=r, dist=dist, sort=False, parallel=parallel
    )

    return within_radius.explode().drop_nulls().value_counts(sort=True, parallel=parallel)


def query_nb_cnt(
    *features: str | pl.Expr,
    r: float | str | pl.Expr | Iterable[float],
    dist: Distance = "sql2",
    parallel: bool = False,
) -> pl.Expr:
    """
    Return the number of neighbors within (<=) radius r for each row under the given distance
    metric. The point itself is always a neighbor of itself.

    Parameters
    ----------
    *features : str | pl.Expr
        Other columns used as features
    r : float | Iterable[float] | pl.Expr | str
        If this is a scalar, then it will run the query with fixed radius for all rows. If
        this is a list, then it must have the same height as the dataframe. If
        this is an expression, it must be an expression representing radius. If this is a str,
        it must be the name of a column
    dist : Literal[`l1`, `l2`, `sql2`, `inf`]
        Note `sql2` stands for squared l2.
    parallel : bool
        Whether to run the distance query in parallel. This is recommended when you
        are running only this expression, and not in group_by() or over() context.
    """
    if dist in ("cosine", "h", "haversine"):
        raise ValueError(f"Distance `{dist}` doesn't work with current implementation.")

    if isinstance(r, (float, int)):
        rad = pl.lit(pl.Series(values=[r], dtype=pl.Float64))
    elif isinstance(r, pl.Expr):
        rad = r
    elif isinstance(r, str):
        rad = pl.col(r)
    else:
        rad = pl.lit(pl.Series(values=r, dtype=pl.Float64))

    return pl_plugin(
        symbol="pl_nb_cnt",
        args=[rad] + [to_expr(x) for x in features],
        kwargs={
            "k": 0,
            "metric": dist,
            "parallel": parallel,
            "skip_eval": False,
            "skip_data": False,
        },
    )
